/* $Id: node.h,v 1.7 2005/12/24 16:56:28 jonathan Exp $
 * Jonathan Ledlie, Harvard University.
 * Copyright 2005.  All rights reserved.
 */

#ifndef NODE_H
#define NODE_H

#include <map>
#include <vector>
#include "point.h"
#include "samples.h"

class Node {
  friend ostream& operator << (ostream& os, Node& n);
 public:
  map<int,Samples*> node2samples;
  double weightedError;
  int lastUpdateTo;
  int lastUpdateFrom;
  int sampleCount;
  int id;
  double distanceDelta;
  vector<double> fixedSample;

  Node ();
  ~Node ();
  Point *vec;
  Point *appVector;
  Point *nearestNeighbor;
  deque<Point*> startCoords;
  deque<Point*> currentCoords;
  int fixedNearestNeighborId;
  double fixedNearestNeighborDistance;

  void Refresh();
  bool isInInitialState ();
  bool hasConfidence ();
  void clearCoords ();
  double getSample (int yourId);
  void addSample 
    (int yourId, double sample, 
     Point *yourCoord, double yourWeightedError,
     Point *yourAppCoord, int stamp);
  bool updateAppVector (double systemDistanceDelta, int stamp,
			double &appComparison);
  void printVector ();
  void addDistanceDelta (double dd);
  double findLocalRelativeDistance (Point *startCentroid,
				    Point *currentCentroid);
  double findEnergyDistance ();
  void addToNearestNeighbor (Point *neighbor, int timestamp);

  void setFixedSample (int yourId, double sample);
  double getFixedSample (int yourId);
  double getFixedSamplePercentile (double p);
  void setFixedSampleArray (int nodeCount);

  void printCoord (FILE *fp);
  void printAppCoord (FILE *fp);
  void printStat (FILE *fp);

  static char appCoordUpdateMethod;
  static uint windowSize;
  static double appCoordUpdateThreshold;
};

#endif
