/* $Id: vivaldi.cc,v 1.21 2006/01/03 23:23:15 jonathan Exp $
 * Jonathan Ledlie, Harvard University.
 * Copyright 2005.  All rights reserved.
 */

/*
correlation between RE and:
- distance
- absolute location
- relative location (vector from me)
- guys in some ball around me have some distribution of REs to a ball around you.
- I have an RE to the guys in a ball around you
My RE to you is X, and my distance to you is Y.  Predict not only distance to Z, but also RE to Z.
- could weight RE prediction based on it's distance's to neighbors (or decreasing expoentially with distance)
- or simply set it equal to nearest neighbor
- could assume that we are told this guys coord _and_ RE (not just his coord) with respect to one of our neighbors.  could then multiply our error to neighbor by this number to get bound.  problem is that our neighbor might not be the one to give us his coord.
- another separate question is, what causes high inaccuracy?
*/

#include "vivaldi.h"
#include "node.h"
#include "new.cc"

#define MAXNODE 10000
#define MAXTIME 100000
#define BASESIZE 150
#define MAX_RANGE 150

#define MAX_NMF_DIM 10

#define TINY 1.0e-10
#define NMAX 500000
#define DIM 7 



int unstructured_link_matrix[1750][1750];

int selected[MAXNODE];
int computed[MAXNODE]; // 2006-10-27
int base_layer_nodes[BASESIZE];
int base_layer_nb[BASESIZE][64];
int period[MAXNODE], last_period[MAXNODE], to_kick[MAXNODE];
int belong_to[MAXNODE];
int belong_to_L2[MAXNODE];
// Output the Load
int output_load[MAXNODE];


  float in_node6[MAXNODE][MAX_NMF_DIM];
  float out_node6[MAXNODE][MAX_NMF_DIM];



// GNP Error of each Node
float min_fit_error2[MAXNODE];
float min_fit_error4[MAXNODE];

// Overhead for each Node
int control_traffic_per_node[MAXNODE];
int distance_probe_per_node[MAXNODE];

// Overhead by time
int control_traffic_by_time[MAXTIME];
int distance_probe_by_time[MAXTIME];


int Prim_table[MAXNODE];
int ESM_table[MAXNODE];
int ESM_thirty_candidate[MAXNODE];

FILE *distribution_relative_fp;
FILE *relative_fp;
FILE *rrl_fp;
FILE *rrl_k_fp;
FILE *cnl_fp;
FILE *nnl_fp;
FILE *ls_fp;
FILE *range_accuracy_fp;
FILE *reverse_range_accuracy_fp;
FILE *output_load_fp;
FILE *min_fit_error_fp; // Analysis the relationship between error & final relative error

int hvivaldi = 0; // hvivaldi = 0, HNC Model, hvivaldi = 1, HVivaldi Model

int vivaldi = 0; //vivaldi = 0, Random Reference Nodes, vivaldi = 1, selective random, vivaldi = 2, selective, cluster
int stable = 1; // stable = 0, ON/OFF Model, stable = 1, Stable Model, stable = 2, Diversity Improvement Model
int on_off = 30;
double average_time = 120; 
//double tempk = 1.5; 		
//double tempk = 1 + ((double)on_off) / (average_time - on_off);
double tempk = log10(20)/log10(average_time/on_off);
//double tempk = 1.0;

#define LANDMARK 32

float **medians = NULL;
int landmark_nodes[MAXNODE][LANDMARK];
int binning[MAXNODE][32];


int  BETA= 4;


int nodeCount = 0;
int neighborCount = 0;
int lineCount = 0;
char *outputFormat = NULL;
int withinMeasurementError = 0;
int maxNeighborCount = 0;
double measurement_error = 0.;
double ERROR_FRACTION = 0.25;
double DAMPENING_FRACTION = 0.25;
int SAMPLE_EXPIRATION = 0;
bool FULL_LATENCY_MATRIX = false;
int FRAME_INTERVAL = 1000;
Node *node = NULL;

Node *node1 = NULL;

Node *node2 = NULL;
Node *node3 = NULL;

Node *node4 = NULL;
Node *node5 = NULL;
Node *node6 = NULL;

char *outputPrefix = NULL;
int lastFrameStamp = 0;

// Vivaldi
int **neighbors = NULL;
// Pharos (2-Level Vivaldi)
int **global_neighbors = NULL;
int **local_neighbors = NULL;
// Noname (3-Level Vivaldi)
int **L0_neighbors = NULL;
int **L1_neighbors = NULL;
int **L2_neighbors = NULL;


FILE *runLogFP = NULL;
FILE *coordFP = NULL;
FILE *reFP = NULL;
FILE *sumreFP = NULL;
FILE *nodeFP = NULL;
FILE *secFP = NULL;

vector<double> sec_sysRawRE_list;
double sec_sysDD_sum;
vector<double> sec_appRawRE_list;
double sec_appDD_sum;

FILE* openOutputFile (char *prefix, char *suffix);
void openOutputFile (FILE *&fp, char *prefix, char *suffix, int stamp);
void closeOutputFile (FILE *fp);
void perSecondOutput (int stamp);
bool isNeighbor (int myId, int yourId);
bool isGlobalNeighbor (int myId, int yourId);
bool isLocalNeighbor (int myId, int yourId);
bool isL0Neighbor (int myId, int yourId);
bool isL1Neighbor (int myId, int yourId);
bool isL2Neighbor (int myId, int yourId);

bool is_base_layer_Neighbor (int mybaseId, int yourbaseId);

int search_base_layer(int yourId);

void MST(int vv, int tree_root_node);
void ESM(int vv, int tree_root_node);
void LGK(int vv, int tree_root_node);
//void SRV_Selection(int);
void SRV(int vv);
void Unstructured(int vv);

double ftol = 0.00001;

int first_beta[32];			// L1 Landmarks
int first_beta_test[32];	// Test set of L1 Landmarks 
int Cluster_Num[32];		// L1 Clusters

/*
int first_beta2[LANDMARK];
int first_beta2_test[LANDMARK];
int Cluster_Num2[LANDMARK];
*/

int first_beta_L2[32]; 			// L2 Landmarks
int first_beta_test_L2[32]; 	// Test set of L2 Landmarks
int Cluster_Num_L2[32];			// L2 Clusters

int base_pic_layer_nodes[32];


double cost_calc() // first_beta[BETA]
{
	int i, j, k, conf;
	int selected[MAXNODE];

	double result = 0.0;
	double TIV_result = 0.0;
	double min_dist;

	int total_triangle = 0;
	int total_TIV = 0;

	for(i=0;i<nodeCount;i++)
	{
		min_dist = 10000;
		conf = 1;
		for(j=0;j<BETA;j++)
		{
			if (medians[i][first_beta_test[j]] < 0)
			{
				continue;
			}
			if (i == first_beta_test[j]) //???
			{
				min_dist = 0;
				conf = first_beta_test[j];
				break;
			}
			if (medians[i][first_beta_test[j]] < min_dist)
			{
				min_dist = medians[i][first_beta_test[j]];
				conf = first_beta_test[j];
			}
		}
		selected[i] = conf;
		// node i belongs to cluster 'Conf' //
		
//		if (conf != -1)
		result += min_dist * min_dist;
		
	}
	/*
	for(i=0;i<nodeCount;i++)
		for(j=i+1;j<nodeCount;j++)
		{
			if (selected[i] != selected[j] || medians[i][j] < 0)
				continue;
			for(k=j+1;k<nodeCount;k++)
			{
				if (selected[i] != selected[k])
					continue;
				if (medians[i][k] < 0 || medians[j][k] < 0)
					continue;
				total_triangle++;
				if (medians[i][j] + medians[j][k] < medians[i][k]
					|| medians[i][k] + medians[k][j] < medians[i][j]
					|| medians[j][i] + medians[i][k] < medians[j][k])
					total_TIV++;
				
			}
		}
		*/
	return result;
//	printf("[%d,%d]", total_TIV, total_triangle);
//	return (double)total_TIV *100/ total_triangle;
}

void Landmark_Hierarchy_Construction()
{
	double gamma;
	if (nodeCount == 1143 || nodeCount == 1740)
		gamma = 50; // King
	else if (nodeCount == 226 || nodeCount == 169)
		gamma = 20; // PlanetLab
	double xita = 90;
	int i, j, k;
	int test;
	int ok;
	int round, turn;
	int selected[MAXNODE];
	int from, to;
	double min_dist, max_dist;

	double quality[BETA];
	int num[BETA];

	for(i=0;i<nodeCount;i++) selected[i] = -1;
	printf("\nForming global cluster...\n");
	i = 0;
	while(1)
	{
		test = rand() % nodeCount;
		printf("%d,%d ", i, test);
		ok = 1;

/*
		for(j=0;j<i;j++)
		{
			// Ever Selected?
			if (test == first_beta[j])
			{
				ok = 0;
				break;
			}
			// Distance above gamma?

//			if (medians[test][first_beta[j]] > -0.5 && medians[test][first_beta[j]] < gamma)
			if (medians[test][first_beta[j]] > -0.5)
			{
				ok = 0;
				break;
			}

		}
*/
		if (selected[test] != -1)
			ok = 0;

		if (ok == 0) continue;
		first_beta[i] = test;
		selected[test] = i;
		i++;
		if (i == BETA) break;
	}
	for(i=0;i<BETA;i++)
	{
		first_beta_test[i] = first_beta[i];
	}
	double now_cost = cost_calc();
	printf("\n");
	printf("0: %.2f\n", now_cost);
	printf("!!!\n");
	for(i=0;i<10000;i++)
	{
		int ok, find_min, belong;
		double min_dist;
//		printf("(%d)", i);	
		k = rand() % nodeCount;
		for(j=0;j<nodeCount;j++)
		{
			// eval : (k+j) % input_node_number
			ok = 1;
			for(find_min=0;find_min<BETA;find_min++)
			{
				if (first_beta[find_min] == (k+j) % nodeCount)
				{
					ok = 0;
					break;
				}
			}
			if (ok == 0) continue;			
			min_dist = 1000000;
			for(find_min=0;find_min<BETA;find_min++)
			{
				if (medians[first_beta[find_min]][(k+j) % nodeCount] < 0) continue;
				if (min_dist>medians[first_beta[find_min]][(k+j) % nodeCount])
				{
					min_dist = medians[first_beta[find_min]][(k+j) % nodeCount];
					belong = find_min;
				}
			}
			for(find_min=0;find_min<BETA;find_min++)				
			{
				if (find_min == belong)
					first_beta_test[find_min] = (k+j) % nodeCount;
				else
					first_beta_test[find_min] = first_beta[find_min];
			}
			if (cost_calc()<now_cost)
			{
				now_cost = cost_calc();
				break;
			}
		}
		if (j == nodeCount)
		{
			printf("Best Result!\n");
			break;
		}
		for(j=0;j<BETA;j++) first_beta[j] = first_beta_test[j];
		printf("[%d]%.2f\t", i+1, now_cost);
	}
//	for(i=0;i<BETA;i++) bin_host[i] = first_beta[i];

	printf("\nL1 Binning: ");
	for(i=0;i<BETA;i++)
		printf("%d ", first_beta[i]);
	printf("Cost: %.2f", now_cost);
	for(i=0;i<BETA;i++) Cluster_Num[i] = 0;
	for(i=0;i<nodeCount;i++)
	{
		min_dist = 1000000;		
		for(j=0;j<BETA;j++)
		{
			if (i == first_beta[j])
			{
				belong_to[i] = j;
				break;
			}
			if (medians[i][first_beta[j]] < 0) continue;
			if (min_dist > medians[i][first_beta[j]])
			{
				min_dist = medians[i][first_beta[j]];
				belong_to[i] = j;
			}
		}
		printf("%d ", belong_to[i]);
		Cluster_Num[belong_to[i]]++;
	}
	printf("\nNode Number for Each Cluster:\n");
// Node Number for Each Cluster:
// 21 72 32 324 2 368 210 1 3 452 6 147 60 1 3 38
	for(i=0;i<BETA;i++) 
		printf("%d ", Cluster_Num[i]);
	printf("\n");
}

/*


double cost_calc_L2() // first_beta_L2[BETA*2]
{
	int i, j, k, conf;
	int selected[MAXNODE];

	double result = 0.0;
	double min_dist;

	for(i=0;i<nodeCount;i++)
	{
		min_dist = 10000;
		conf = 1;
		for(j=0;j<BETA*2;j++)
		{
			if (medians[i][first_beta_test_L2[j]] < 0)
			{
				continue;
			}
			if (i == first_beta_test_L2[j]) //???
			{
				conf = -1;
				break;
			}
			if (medians[i][first_beta_test_L2[j]] < min_dist)
			{
				min_dist = medians[i][first_beta_test_L2[j]];
				conf = j;
			}
		}
		selected[i] = conf;
		if (conf != -1)
			result += min_dist * min_dist;
		
	}
	return result;
}


void Landmark_Hierarchy_Construction_L2()
{
	double gamma;
	if (nodeCount == 1143 || nodeCount == 1740)
		gamma = 50; // King
	else if (nodeCount == 226 || nodeCount == 169)
		gamma = 20; // PlanetLab
	gamma = 20;
	double xita = 90;
	int i, j, k;
	int test;
	int ok;
	int round, turn;
	int selected[MAXNODE];
	int from, to;
	double min_dist, max_dist;

//	double quality[BETA*2];
//	int num[BETA*2];

	for(i=0;i<nodeCount;i++) selected[i] = -1;
	printf("\nForming global cluster...\n");
	i = 0;
	while(1)
	{
		test = rand() % nodeCount;
		ok = 1;
		for(j=0;j<i;j++)
		{
			// Ever Selected?
			if (test == first_beta_L2[j])
			{
				ok = 0;
				break;
			}
			// Distance above gamma?

			if (medians[test][first_beta_L2[j]] > -0.5 && medians[test][first_beta_L2[j]] < gamma)
			{
				ok = 0;
				break;
			}

		}
		if (ok == 0) continue;
		first_beta_L2[i] = test;
		selected[test] = i;
		i++;
		if (i == BETA*2) break;
	}
	for(i=0;i<BETA*2;i++)
	{
		first_beta_test_L2[i] = first_beta_L2[i];
	}
	double now_cost = cost_calc_L2();
	printf("\n");
//	printf("0: %.2f\n", now_cost);
	for(i=0;i<50000;i++)
	{
		int ok, find_min, belong;
		double min_dist;
		
		k = rand() % nodeCount;
		for(j=0;j<nodeCount;j++)
		{
			// eval : (k+j) % input_node_number
			ok = 1;
			for(find_min=0;find_min<BETA*2;find_min++)
			{
				if (first_beta_L2[find_min] == (k+j) % nodeCount)
				{
					ok = 0;
					break;
				}
			}
			if (ok == 0) continue;			
			min_dist = 1000000;
			for(find_min=0;find_min<BETA*2;find_min++)
			{
				if (medians[first_beta_L2[find_min]][(k+j) % nodeCount] < 0) continue;
				if (min_dist>medians[first_beta_L2[find_min]][(k+j) % nodeCount])
				{
					min_dist = medians[first_beta_L2[find_min]][(k+j) % nodeCount];
					belong = find_min;
				}
			}
			for(find_min=0;find_min<BETA*2;find_min++)				
			{
				if (find_min == belong)
					first_beta_test_L2[find_min] = (k+j) % nodeCount;
				else
					first_beta_test_L2[find_min] = first_beta_L2[find_min];
			}
			if (cost_calc_L2()<now_cost)
			{
				now_cost = cost_calc_L2();
				break;
			}
		}
		if (j == nodeCount)
		{
			printf("Best Result!\n");
			break;
		}
		for(j=0;j<BETA*2;j++) first_beta_L2[j] = first_beta_test_L2[j];
		printf("[%d]%.2f\t", i+1, now_cost);
	}
//	for(i=0;i<BETA;i++) bin_host[i] = first_beta[i];

	printf("\nL2 Binning: ");
	for(i=0;i<BETA*2;i++)
		printf("%d ", first_beta_L2[i]);
	printf("Cost: %.2f", now_cost);
	for(i=0;i<BETA*2;i++) Cluster_Num_L2[i] = 0;
	for(i=0;i<nodeCount;i++)
	{
		min_dist = 1000000;		
		for(j=0;j<BETA*2;j++)
		{
			if (i == first_beta_L2[j])
			{
				belong_to_L2[i] = j;
				break;
			}
			if (medians[i][first_beta_L2[j]] < 0) continue;
			if (min_dist > medians[i][first_beta_L2[j]])
			{
				min_dist = medians[i][first_beta_L2[j]];
				belong_to_L2[i] = j;
			}
		}
		printf("%d ", belong_to_L2[i]);
		Cluster_Num_L2[belong_to_L2[i]]++;
	}
	printf("\nNode Number for Each Cluster:\n");
// Node Number for Each Cluster:
// 21 72 32 324 2 368 210 1 3 452 6 147 60 1 3 38
	for(i=0;i<BETA*2;i++) 
		printf("%d ", Cluster_Num_L2[i]);
	printf("\n");
}
*/

void processSample2 (int o_stamp, int myId, int o_yourId, float o_rawLatency) {

  // o_ refers to this observation
  // s_ refers to the sample being processed,
  //    as we go through our samples

  // distance refers to metric space distance
  // latency to actual observations

  if (debug)
    cout << "\n\nLINE: "<< o_stamp << " me " << myId << " you " 
	 << o_yourId << " s " << o_rawLatency << endl;

  if (myId >= nodeCount || o_yourId >= nodeCount) {
    printf ("id out of range, mine %d, yours %d\n", myId, o_yourId);
    exit (-1);
  }

  // remember samples from this node
//  node[myId].addSample
//    (o_yourId, o_rawLatency, node[o_yourId].vec, node[o_yourId].weightedError,
//     node[o_yourId].appVector, o_stamp);

  // retrieve our idea of the underlying latency for this link
//  double o_latency = node[myId].getSample (o_yourId);
  float o_latency = o_rawLatency;

  // remember when we last exchanged info with this guy
  node[myId].lastUpdateFrom = o_stamp;
  node[o_yourId].lastUpdateTo = o_stamp;

  if (o_latency > 0) {

    if (debug)
      cout << "my coord" << node[myId].vec << endl;

    double o_distance = node[myId].vec->getEuclideanDistance (node[o_yourId].vec);
    while (o_distance == 0) {
      node[myId].vec->bump();
      o_distance = node[myId].vec->getEuclideanDistance (node[o_yourId].vec);
    }
    if (o_distance == 0) {
      cout << "my coord " << node[myId].vec
	   << "your coord " << node[o_yourId].vec << endl;
    }
    assert (!isnan(o_distance));

    double o_relativeError = fabs(o_distance - o_latency) / o_latency;
    assert (!isnan(o_relativeError));

    double o_rawRelativeError = fabs(o_distance - o_rawLatency) / o_rawLatency;
    assert (!isnan(o_rawRelativeError));

    // EMWA on weightedError, based on how much we trust this current sample
    // update our confidence based on the confidence in this sample
    assert (node[myId].weightedError + node[o_yourId].weightedError > 0);

    double o_weight = node[myId].weightedError /
      (node[myId].weightedError + node[o_yourId].weightedError);
    assert (!isnan(o_weight));

    double o_alphaWeightedError = ERROR_FRACTION * o_weight;
    assert (!isnan(o_alphaWeightedError));

    // update our confidence
    node[myId].weightedError = (o_relativeError * o_alphaWeightedError) +  (node[myId].weightedError * (1. - (o_alphaWeightedError)));

    if (debug) {
      printf ("weightedError = %f\n", node[myId].weightedError);
    }
    if (node[myId].weightedError > 1.) node[myId].weightedError = 1.;
    if (node[myId].weightedError < 0.) node[myId].weightedError = 0.;

    assert (!isnan(node[myId].weightedError));

    Point o_force;

    // find direction
	Point *o_unitVector = node[myId].vec->getDirection (node[o_yourId].vec);
	if (o_unitVector == NULL) {
	  o_unitVector = Point::getRandomUnitVector();
	}

	if (debug)
	  cout << "unitVector to id "<< o_yourId << " = "<<o_unitVector <<endl;

	if (debug) {
	  printf ("sample weighted error = %f\n", o_weight);
	  printf ("predicted distance = %f\n", o_distance);
	  printf ("actual distance =	  %f\n", o_latency);
	}

	double o_error = o_latency - o_distance;
	
	// delta
	double o_dampening = DAMPENING_FRACTION * o_weight;

	o_unitVector->scale (o_error * o_dampening);

	if (debug)
	  cout << "scaled unitVector "<<o_unitVector <<endl;
	o_force.add (o_unitVector);


    if (debug) {
      cout << "before update "<< node[myId].vec << endl;
      cout << "force " << &o_force << endl;
    }

    node[myId].vec->add(&o_force);
    node[myId].vec->checkHeight();

    if (debug) {
      cout << "after update "<< node[myId].vec << endl;
    }

	delete o_unitVector;

	
    double o_distanceDelta = fabs(o_force.length());
    node[myId].addDistanceDelta (o_distanceDelta);

    //assert (o_distanceDelta < 500);
    if (o_distanceDelta > 500) {
      printf ("problem\n");
    }

    if (debug) {
      cout << "dd " << o_distanceDelta << endl;
    }

    if (debug) {
      double o_distance = node[myId].vec->getEuclideanDistance (node[o_yourId].vec);
      cout << "new dist " << o_distance << endl;
    }



  if (debug) {
    printf ("\n\n\n");
  }
  	}
}


void processSample (int o_stamp, int myId, int o_yourId, float o_rawLatency) {

  // o_ refers to this observation
  // s_ refers to the sample being processed,
  //    as we go through our samples

  // distance refers to metric space distance
  // latency to actual observations

  if (debug)
    cout << "\n\nLINE: "<< o_stamp << " me " << myId << " you " 
	 << o_yourId << " s " << o_rawLatency << endl;

  if (myId >= nodeCount || o_yourId >= nodeCount) {
    printf ("id out of range, mine %d, yours %d\n", myId, o_yourId);
    exit (-1);
  }

  // remember samples from this node
  node[myId].addSample
    (o_yourId, o_rawLatency, node[o_yourId].vec, node[o_yourId].weightedError,
     node[o_yourId].appVector, o_stamp);

  // retrieve our idea of the underlying latency for this link
  double o_latency = node[myId].getSample (o_yourId);

  // remember when we last exchanged info with this guy
  node[myId].lastUpdateFrom = o_stamp;
  node[o_yourId].lastUpdateTo = o_stamp;

  if (o_latency > 0) {

    if (debug)
      cout << "my coord" << node[myId].vec << endl;

    double o_distance = node[myId].vec->getEuclideanDistance (node[o_yourId].vec);
    while (o_distance == 0) {
      node[myId].vec->bump();
      o_distance = node[myId].vec->getEuclideanDistance (node[o_yourId].vec);
    }
    if (o_distance == 0) {
      cout << "my coord " << node[myId].vec
	   << "your coord " << node[o_yourId].vec << endl;
    }
    assert (!isnan(o_distance));

    double o_relativeError = fabs(o_distance - o_latency) / o_latency;
    assert (!isnan(o_relativeError));

    double o_rawRelativeError = fabs(o_distance - o_rawLatency) / o_rawLatency;
    assert (!isnan(o_rawRelativeError));

    // EMWA on weightedError, based on how much we trust this current sample
    // update our confidence based on the confidence in this sample
    assert (node[myId].weightedError + node[o_yourId].weightedError > 0);

    double o_weight = node[myId].weightedError /
      (node[myId].weightedError + node[o_yourId].weightedError);
    assert (!isnan(o_weight));

    double o_alphaWeightedError = ERROR_FRACTION * o_weight;
    assert (!isnan(o_alphaWeightedError));

    // update our confidence
    node[myId].weightedError = (o_relativeError * o_alphaWeightedError) +
      (node[myId].weightedError * (1. - (o_alphaWeightedError)));

    if (debug) {
      printf ("weightedError = %f\n", node[myId].weightedError);
    }
    if (node[myId].weightedError > 1.) node[myId].weightedError = 1.;
    if (node[myId].weightedError < 0.) node[myId].weightedError = 0.;

    assert (!isnan(node[myId].weightedError));

    Point o_force;
    int measurementsUsed = 0;
    int o_oldestSample = o_stamp;

    // randomize processing of our neighbors
    // when SAMPLE_EXPIRATION == 0, we only process the most recent sample

    vector<int> validNodes;
    for (map<int,Samples*>::iterator samplesIter =
           node[myId].node2samples.begin(); 
         samplesIter != node[myId].node2samples.end(); samplesIter++) {

      int s_yourId = samplesIter->first;
      Samples *sample = samplesIter->second;


  //  Valid id 83 sample [sC [0 0 0] aC [0 0 0] wE 1 s 47.324 ]
      if (sample->stamp < o_stamp - SAMPLE_EXPIRATION &&
	  SAMPLE_EXPIRATION >= 0) {
        if (debug)
          cout << "Expired id " << s_yourId << " sample " << sample << endl;
      } else {
        if (debug)
          cout << "Valid id " << s_yourId << " sample " << sample << endl;
        validNodes.push_back (s_yourId);
	if (o_oldestSample > sample->stamp) {
	  o_oldestSample = sample->stamp;
	}
      }
    }

    // this is unnecessary
    random_shuffle(validNodes.begin(),
                   validNodes.end());
    double o_sampleWeightSum = 0.;

    // let o_sampleWeightSum be sum of
    // relative age of each sample.
    // Then weigh each sample by its weight/o_sampleWeightSum
    for (int validNodeIndex = 0; validNodeIndex < validNodes.size();
         validNodeIndex++) {
      int s_yourId = validNodes[validNodeIndex];
      Samples *sample = node[myId].node2samples[s_yourId];
      o_sampleWeightSum += sample->stamp - o_oldestSample;
    }
    assert (o_sampleWeightSum >= 0);

    for (int validNodeIndex = 0; validNodeIndex < validNodes.size();
         validNodeIndex++) {

      int s_yourId = validNodes[validNodeIndex];
      Samples *sample = node[myId].node2samples[s_yourId];

      double s_distance = node[myId].vec->getEuclideanDistance
        (sample->vec);
      while (s_distance == 0.) {
	node[myId].vec->bump();
	s_distance = node[myId].vec->getEuclideanDistance (sample->vec);
      }
      assert (!isnan(o_distance));

      // find direction
      Point *s_unitVector = node[myId].vec->getDirection (sample->vec);
      if (s_unitVector == NULL) {
        s_unitVector = Point::getRandomUnitVector();
      }

      double s_latency = node[myId].getSample (s_yourId);
      if (debug)
        cout << "unitVector to id "<< s_yourId << " = "<<s_unitVector <<endl;

      // how much to weigh this sample
      // are his coords more reliable than mine?

      if (!(node[myId].weightedError + sample->weightedError > 0.)) {
	printf ("me %d you %d meas %d myWE %f yourWE %f\n",
		myId, s_yourId, measurementsUsed, node[myId].weightedError,
		sample->weightedError);
	exit (0);
      }

      assert (node[myId].weightedError + sample->weightedError > 0.);

      double s_weight = node[myId].weightedError /
        (node[myId].weightedError + sample->weightedError);


      if (debug) {
        printf ("sample weighted error = %f\n", s_weight);
        printf ("predicted distance = %f\n", s_distance);
        printf ("smoothed distance =    %f\n", s_latency);
      }

      // Meant to account for blips in very low latency links
      // When measurement_error == 0, this doesn't do anything

      // e_{s}
      if ( (s_distance < s_latency &&
            s_distance + measurement_error >= s_latency) ||
           (s_distance > s_latency &&
            s_distance - measurement_error <= s_latency)) {
        printf ("within measurement error\n");
        s_distance = s_latency;
        withinMeasurementError++;
      }


      // error of sample
      double s_error = s_latency - s_distance;
    
      // delta
      double s_dampening = DAMPENING_FRACTION * s_weight;
      if (debug) {
        printf ("dampening = %f for vector adjustment\n", s_dampening);
      }

      double s_sampleWeight = 1.;
      if (o_sampleWeightSum > 0) {
	s_sampleWeight = (sample->stamp - o_oldestSample)/o_sampleWeightSum;
      }

      s_unitVector->scale (s_error * s_dampening * s_sampleWeight);
      if (debug)
        cout << "scaled unitVector "<<s_unitVector <<endl;
      o_force.add (s_unitVector);
      if (debug)
        cout << "total force "<<&o_force <<endl;
      delete s_unitVector;
      measurementsUsed++;
    }
    
    // Update the local coordinate

    if (debug) {
      cout << "before update "<< node[myId].vec << endl;
      cout << "force " << &o_force << endl;
    }

    node[myId].vec->add(&o_force);
    node[myId].vec->checkHeight();

    if (debug) {
      cout << "after update "<< node[myId].vec << endl;
    }

    double o_distanceDelta = fabs(o_force.length());
    node[myId].addDistanceDelta (o_distanceDelta);

    //assert (o_distanceDelta < 500);
    if (o_distanceDelta > 500) {
      printf ("problem\n");
    }

    if (debug) {
      cout << "dd " << o_distanceDelta << endl;
    }

    if (debug) {
      double o_distance = node[myId].vec->getEuclideanDistance (node[o_yourId].vec);
      cout << "new dist " << o_distance << endl;
    }

    double o_appCoordDistance = node[myId].appVector->getEuclideanDistance
      (node[o_yourId].appVector);
    Point *o_oldAppPosition = new Point (node[myId].appVector);

    // update nearest neighbor after changing our coordinate,
    // and before thinking about updating the appVector
    node[myId].addToNearestNeighbor (node[o_yourId].vec, o_stamp);

    double o_appComparison = 0.;
    bool o_updatedAppVector = node[myId].updateAppVector
      (o_distanceDelta, o_stamp,
       o_appComparison);

    if (debug)
      cout << "appCoordDistance " << o_appCoordDistance << endl;
    double o_appDistanceDelta =
      o_oldAppPosition->getAbsoluteDistance (node[myId].appVector);
    double o_appRawRelativeError = 
      fabs(o_appCoordDistance - o_rawLatency) / o_rawLatency;   
    double o_appRelativeError = 
      fabs(o_appCoordDistance - o_latency) / o_latency;   
    delete o_oldAppPosition;

    if (secFP) {
      sec_sysRawRE_list.push_back (o_rawRelativeError);
      sec_sysDD_sum += o_distanceDelta;
      sec_appRawRE_list.push_back (o_appRawRelativeError);
      sec_appDD_sum += o_appDistanceDelta;
    }

    //if (outputFormat == 'o' || outputFormat == 'c') {
    if (runLogFP) {
      fprintf (runLogFP, "%d me %3d you %3d s %7.2f ss %7.2f confi %4.3f dd %6.3f re %5.3f sA %d ",
               o_stamp, myId, o_yourId, o_rawLatency, o_latency,
               node[myId].weightedError, o_distanceDelta,
               o_relativeError, o_oldestSample);
      /*
        printf ("appUp ");
	if (updatedAppVector)
        printf ("1 ");
	else
        printf ("0 ");
	printf ("appDD %8.4f rAppRE %8.4f sAppRE %8.4f appComp %8.4f",
        appDistanceDelta, 
        appRawRelativeError,appRelativeError,
        appComparison);
      */
      //if (outputFormat == 'c') {
      //printf (" [ ");
      //node[myId].printVector ();
      //printf (" ]");
      //}
      fprintf (runLogFP, " [ ");
      node[myId].printCoord(runLogFP);
      fprintf (runLogFP, " ]\n");
    }
  }

  if (debug) {
    printf ("\n\n\n");
  }
}

int check_load(int node_id)
{
	int i, j, total;
	for(i=0;i<nodeCount;i++)
	{
		for(j=0;j<neighborCount;j++)
		{
			if (neighbors[i][j] == node_id)
				total++;
		}
	}		
	return total;
}

void printUsage () {

  cout << "vivaldi\n"
       << " -n nodeCount\n"
       << " -f sampleFile\n"
       << " -e error fraction (= " << ERROR_FRACTION << ")\n"
       << " -u dampening fraction (= " << DAMPENING_FRACTION << ")\n"
       << " -o measurement error (overhead) (= "<< measurement_error <<"ms)\n"
       << " -h ping history count (sample memory per node) (= "<<PING_HISTORY_COUNT<<")\n"
       << "    -1 uses an EWMA instead of a moving percentile filter\n"
       << "    and -p as alpha\n"
       << " -i min ping history count before using sample (= "<<MIN_HISTORY_COUNT<<")\n"
       << " -p percentile to use in moving percentile filter (0..1) (="<<PING_SAMPLE_PERCENTILE<<")\n"
       << " -g [ocsrlap] specialize output to... \n"
       << "   o: change with each observation\n"
       << "   c: coords at end of run\n"
       << "   n: node statistics\n"
       << "   r: print relative error using median link latency file\n"
       << "   R: print summary of relative error using median link latency file\n"
       << "   l: print relative rank loss using median link latency file\n"
       << "   a: print normalized absolute rank loss using median link latency file\n"
       << "   p: relative application latency penalty\n"
       << "   g: print coords every second (for gnuplot)\n"
       << "   s: print per-second summary\n"
       << " -k neighbor count (defaults to all)\n"
       << " -x dimensions (= "<<DIMENSIONS<< ")\n"
       << " -d [debug] ("<<debug<<")\n"
       << " -m [e|l|s|a] ("<<Node::appCoordUpdateMethod<<"\n"
       << "   e: measure energy between two windows\n"
       << "   l: local relative distance between centroid of two windows\n"
       << "   s: measure change in system-level coordinate\n"
       << "   a: measure change in application-level coordinate\n"
       << "   c: measure change in application-level coordinate, \n"
       << "      and update it to centroid\n"
       << " -t threshold for updating application-level coordinate (="<<Node::appCoordUpdateThreshold<<")\n"
       << " -w window size (="<<Node::windowSize<<")\n"
       << " -l true latency matrix\n"
       << " -r rounds, choose randomly from latency matrix file\n"

       << " -b true latency matrix is full, e.g. allows asymetric links (=false)\n"

       << " -j sample expiration (in rounds or ms), if negative, no expiration\n"
       << " -O output prefix\n"
       << " -s seed\n"
       << " -v Type of Method (v = 0, HNC, v = 1, HNC(s), v = 2, Vivaldi";
  exit (-1);
}

int already_landmark(int test_node, int target)
{
	int i;
	for(i=0;i<LANDMARK;i++)
	{
		if (landmark_nodes[test_node][i] == target)
			return 1;
	}
	return -1;
}

/*
int already_first_beta(int test_node)
{
	int i;
	for(i=0;i<LANDMARK;i++)
	{
		if (first_beta[i] == test_node)
			return 1;
	}
	return -1;
}
*/

int main (int argc, char **argv) {
  char c =  0;
  char *sampleFile = NULL;
  char *trueLatencyMatrixFile = NULL;
  int seed = getpid();
  int ROUNDS = 0;

  FILE *fp;
  while ((c = getopt(argc,argv,"bf:n:e:u:s:o:x:h:p:dg:m:t:w:i:l:zr:j:O:k:v:c:")) >= 0) {
    switch (c) {
    case 'O':
      outputPrefix = optarg;
      break;
    case 'f':
      sampleFile = optarg;
      break;
    case 'l':
      trueLatencyMatrixFile = optarg;
      break;
    case 'k':
      neighborCount = atoi(optarg);
      break;
    case 'z':
      USE_HEIGHT = true;
      break;
    case 'b':
      FULL_LATENCY_MATRIX = true;
      break;
    case 'j':
      SAMPLE_EXPIRATION = atoi (optarg);
      break;
    case 'h':
      PING_HISTORY_COUNT = atoi(optarg);
      break;
    case 'i':
      MIN_HISTORY_COUNT = atoi(optarg);
      break;
    case 'r':
      ROUNDS = atoi(optarg);
      break;
    case 'n':
      nodeCount = atoi(optarg);
      break;
    case 's':
      seed = atoi(optarg);
      break;
    case 'x':
      DIMENSIONS = atoi(optarg);
      break;
    case 'g':
      outputFormat = optarg;
      break;
    case 'm':
      Node::appCoordUpdateMethod = *optarg;
      break;
    case 'w':
      Node::windowSize = atoi(optarg);
      break;
    case 't':
      Node::appCoordUpdateThreshold = atof(optarg);
      //cout << "here Node::appCoordUpdateThreshold = " << Node::appCoordUpdateThreshold << endl;
      break;
    case 'd':
      debug = true;
      break;
    case 'e':
      ERROR_FRACTION = atof(optarg);
      break;
    case 'p':
      PING_SAMPLE_PERCENTILE = atof(optarg);
      break;
    case 'o':
      measurement_error = atof(optarg);
      break;
    case 'u':
      DAMPENING_FRACTION = atof(optarg);
      break;
    case 'v':
	vivaldi = atoi(optarg);
	break;
	case 'c':
		BETA = atoi(optarg);
		break;
    default:
      printUsage ();
      break;
    }
  }

  srandom (seed);
  SRand (seed);

  if (outputPrefix == NULL) {
    printf("Output file prefix is required.\n");
    printUsage ();
  }

  if (outputFormat == NULL) {
    printf("Output format is required.\n");
    printUsage ();
  }

  if (nodeCount <= 0) {
    printf ("Bad sample file or node count\n");
    printUsage ();
  }

  if (neighborCount <= 0) {
    printf ("Invalid neighbor count: %d\n", neighborCount);
    printUsage ();
  }

  int myId, yourId;
  int stamp = 0;
  float rawLatencySample; 

  int local_nb_count[MAXNODE];
  int L0_nb_count[MAXNODE];
  int L1_nb_count[MAXNODE];
  int L2_nb_count[MAXNODE];

  int global_nb_upbound = neighborCount /2;
  int local_nb_upbound = neighborCount - global_nb_upbound;

  // parse medians file, if available
  if (trueLatencyMatrixFile != NULL) {

    if ((fp = fopen (trueLatencyMatrixFile, "r")) == NULL) {
      printf ("Cannot open file %s", trueLatencyMatrixFile);
      exit (-1);
    }

    medians = new float* [nodeCount];
    neighbors = new int* [nodeCount];

    global_neighbors = new int* [nodeCount];
    local_neighbors = new int* [nodeCount];


    L0_neighbors = new int* [nodeCount];
    L1_neighbors = new int* [nodeCount];
    L2_neighbors = new int* [nodeCount];


    // these temp variables stop compiler from complaining
    for (int myIdT = 0; myIdT < nodeCount; myIdT++) {
      medians[myIdT] = new float[nodeCount];

	  neighbors[myIdT] = new int[neighborCount];

	  global_neighbors[myIdT] = new int[global_nb_upbound];
      local_neighbors[myIdT] = new int[local_nb_upbound];

	  L0_neighbors[myIdT] = new int[(neighborCount/3)];
	  L1_neighbors[myIdT] = new int[(neighborCount/3)];
	  L2_neighbors[myIdT] = new int[(neighborCount/3)];
	  

	  for (int yourIdT = 0; yourIdT < nodeCount; yourIdT++) {
		medians[myIdT][yourIdT] = -1; /////////////// 0 -> -1
      }
	  // Vivaldi
      for (int yourIdT = 0; yourIdT < neighborCount; yourIdT++) {
		neighbors[myIdT][yourIdT] = -1;
      }
	  // Pharos
      for (int yourIdT = 0; yourIdT < global_nb_upbound; yourIdT++) {
		global_neighbors[myIdT][yourIdT] = -1;
      }
      for (int yourIdT = 0; yourIdT < local_nb_upbound; yourIdT++) {

		local_neighbors[myIdT][yourIdT] = -1;
      	}
	  // Pharos++
      for (int yourIdT = 0; yourIdT < neighborCount/3; yourIdT++) {
		L0_neighbors[myIdT][yourIdT] = -1;
		L1_neighbors[myIdT][yourIdT] = -1;
		L2_neighbors[myIdT][yourIdT] = -1;
      }

    }

    while (fscanf (fp, "%d %d %f\n", &myId, &yourId, &rawLatencySample) >= 0) {   
//		printf("%f ", rawLatencySample);
//      assert (rawLatencySample >= 0.);
////////////////////////////////////////////////////
	if (rawLatencySample >= 0)
		{
      medians[myId][yourId] = rawLatencySample;
      if (!FULL_LATENCY_MATRIX)
		medians[yourId][myId] = rawLatencySample;
		}
    }
    fclose (fp);



//	FILE *matrix_king;
//	matrix_king = fopen("king_matrix.txt", "w");
//	for(int ii=0;ii<nodeCount;ii++)
//	{
//		for(int jj=0;jj<nodeCount;jj++)
//		{
//			if (ii == jj)
//				fprintf(matrix_king, "0 ");
//			else
//				fprintf(matrix_king, "%f ", medians[ii][jj]);
//		}
//		fprintf(matrix_king, "\n");
//	}
//	fclose(matrix_king);
//	exit(1);
	Landmark_Hierarchy_Construction();
//	Landmark_Hierarchy_Construction_L2();


    printf("\nAdd nbs\n");
    for (int myId = 0; myId < nodeCount; myId++) {
      // verify that everybody has enough neighbors
      int myNeighborCount = 0;


//     int local_NeighborCount=0;
	 
      for (int yourId = 0; yourId < nodeCount; yourId++) {
		if (medians[myId][yourId] > 0. && yourId != myId) {
			myNeighborCount++;
		}
//		belong_to[MAXNODE];
      }
//	printf("[%d %d]\n", myNeighborCount, neighborCount);
      //assert (myNeighborCount >= neighborCount);

	// My Neighbor Count >= neighborCount	
	if (myNeighborCount >= neighborCount)
	{
	      for (int neighborId = 0; neighborId < neighborCount; neighborId++) {
			int yourId = myId;
			while (yourId == myId || medians[myId][yourId] < -0.5
		       || isNeighbor(myId,yourId))
		       { // <= 0. to < -0.5
				yourId = unifRand (0, nodeCount);
			}
			neighbors[myId][neighborId] = yourId;
			//printf ("me %d nIndex %d you %d\n", myId, neighborId, yourId);			
	      }
	      for (int neighborId = 0; neighborId < global_nb_upbound; neighborId++) {
			int yourId = myId;
			while (yourId == myId || medians[myId][yourId] < -0.5
		       || isGlobalNeighbor(myId,yourId) 
		      || belong_to[myId] == belong_to[yourId])
		       { // <= 0. to < -0.5
				yourId = unifRand (0, nodeCount);
			}
			global_neighbors[myId][neighborId] = yourId;
			//printf ("me %d nIndex %d you %d\n", myId, neighborId, yourId);			
	      }
		/*
	      for (int neighborId = 0; neighborId < (neighborCount/3); neighborId++) {
			int yourId = myId;
			while (yourId == myId || medians[myId][yourId] < -0.5
		       || isL0Neighbor(myId,yourId))
		       { // <= 0. to < -0.5
				yourId = unifRand (0, nodeCount);
			}
			L0_neighbors[myId][neighborId] = yourId;
			//printf ("me %d nIndex %d you %d\n", myId, neighborId, yourId);			
	      }		  
		  L0_nb_count[myId] = (neighborCount /3);
		  */
	}
	else
	{
		printf("\nWarning\n");
		// Add : My Neighbor Count < neigborCount
		int neighborId = 0;
		for(int yourId = 0; yourId < nodeCount; yourId++)
		{
			if (medians[myId][yourId] > -0.5) {
				neighbors[myId][neighborId] = yourId;
				neighborId++;
			}		
		}
	}
	
    }
    printf("Done!");

  } else if (sampleFile == NULL) {
    printf ("Bad sample file or node count\n");
    printUsage ();
  }


  if (PING_SAMPLE_PERCENTILE < 0 || PING_SAMPLE_PERCENTILE > 1) {
    printf ("Bad ping sample percentile\n");
    printUsage ();
  }

  /*
  if (SAMPLE_EXPIRATION < 0)  {
    printf ("Sample expiration must be greater or equal to 0\n");
    printUsage ();
  }
  */

  if (PING_HISTORY_COUNT != -1 && 
      PING_HISTORY_COUNT < 0) {
    printf ("Bad ping history count\n");
    printUsage ();
  }

  if ((Node::appCoordUpdateMethod == 'l' ||
       Node::appCoordUpdateMethod == 'e') &&
      Node::windowSize <= 0) {
    printUsage ();
  }

  //printf ("len %d p %s\n", strlen (outputFormat), outputFormat);

  if (strchr(outputFormat,'o') != NULL)
    runLogFP = openOutputFile (outputPrefix, ".log");
  if (strchr(outputFormat,'c') != NULL)
    coordFP = openOutputFile (outputPrefix, ".coord");
  if (strchr(outputFormat,'r') != NULL)
    reFP = openOutputFile (outputPrefix, ".re");
  if (strchr(outputFormat,'R') != NULL)
    sumreFP = openOutputFile (outputPrefix, ".sumre");
  if (strchr(outputFormat,'n') != NULL)
    nodeFP = openOutputFile (outputPrefix, ".node");
  if (strchr(outputFormat,'s') != NULL)
    secFP = openOutputFile (outputPrefix, "sec.log");

  // Finished checking user input
  // Start processing either the median file
  // or a ping trace

 /////////////////////////////////////////////////////////////////
  node = new Node[nodeCount];	// Basic
  
  node1 = new Node[nodeCount];	// Pure Vivaldi NC

  node2 = new Node[nodeCount];	// HVivaldi -- Global NC
  node3 = new Node[nodeCount];	// HVivaldi -- Local NC

  node4 = new Node[nodeCount];	// L0 NC 
  node5 = new Node[nodeCount];	// L1 NC 
  node6 = new Node[nodeCount];	// L2 NC 


/////////////////////////////////////////////////////////////////

  if (sampleFile != NULL) {
    if ((fp = fopen (sampleFile, "r")) == NULL) {
      printf ("Cannot open file %s", sampleFile);
    }

    while (fscanf (fp, "%d %d %d %f\n", &stamp, &myId, &yourId,
                   &rawLatencySample) > 0) { 

      //if (stamp % 100) {
	//fprintf (stderr, "stamp %d\n", stamp);
      //}

      processSample (stamp, myId, yourId, rawLatencySample);
      perSecondOutput (stamp);
    }
    fclose (fp);

  } else 
  {
    // process from medians

	 //cout << "start" << endl;
	int bad = 0;
	int generation = 0;
	int ii, jj, kk;	

	int load[MAXNODE], age[MAXNODE];


	int base_layer_load[BASESIZE];




	// Add Local NBs


	printf("\nAdd Local nbs\n");
	for (int my_Id = 0; my_Id < nodeCount; my_Id++) {
      // verify that everybody has enough neighbors
	     int local_NeighborCount=0;
	 
	      for (int yourId = 0; yourId < nodeCount; yourId++) 
			if (medians[my_Id][yourId] > -0.5 && yourId != my_Id && belong_to[my_Id] == belong_to[yourId])
					local_NeighborCount++;

			printf("[%d,%d]",my_Id,  local_NeighborCount);
			if (local_NeighborCount >= local_nb_upbound)
				local_nb_count[my_Id] = local_nb_upbound;
			else
				local_nb_count[my_Id] = local_NeighborCount;
		if (local_NeighborCount >  local_nb_upbound)
		{
		      for (int neighborId = 0; neighborId < (local_nb_upbound); neighborId++) {
				int yourId = my_Id;
				while (yourId == my_Id || medians[my_Id][yourId] < -0.5
			       || isLocalNeighbor(my_Id,yourId) || belong_to[my_Id] != belong_to[yourId])
			       { // <= 0. to < -0.5
					yourId = unifRand (0, nodeCount);
				}
				local_neighbors[my_Id][neighborId] = yourId;
				//printf ("me %d nIndex %d you %d\n", myId, neighborId, yourId);			
		      }
		}
		else
		{
			printf("Warning ");
			// Add : My Neighbor Count < neigborCount
			int neighborId = 0;
			for(int yourId = 0; yourId < nodeCount; yourId++)
			{
				if (medians[my_Id][yourId] > -0.5 &&  belong_to[my_Id] == belong_to[yourId]) {
					local_neighbors[my_Id][neighborId] = yourId;
					neighborId++;
				}		
			}
		}
	
    }

/*
	printf("\nAdd L1 nbs\n");

	for (int my_Id = 0; my_Id < nodeCount; my_Id++) {
      // verify that everybody has enough neighbors
	     int L1_NeighborCount=0;
	 
	      for (int yourId = 0; yourId < nodeCount; yourId++) 
			if (medians[my_Id][yourId] > -0.5 && yourId != my_Id && belong_to[my_Id] == belong_to[yourId])
					L1_NeighborCount++;

			printf("[%d,%d]",my_Id,  L1_NeighborCount);
			if (L1_NeighborCount >= neighborCount / 3)
				L1_nb_count[my_Id] = neighborCount / 3;
			else
				L1_nb_count[my_Id] = L1_NeighborCount;
		if (L1_NeighborCount >  neighborCount / 3)
		{
		      for (int neighborId = 0; neighborId < (neighborCount/3); neighborId++) {
				int yourId = my_Id;
				while (yourId == my_Id || medians[my_Id][yourId] < -0.5
			       || isL1Neighbor(my_Id,yourId) || belong_to[my_Id] != belong_to[yourId])
			       { // <= 0. to < -0.5
					yourId = unifRand (0, nodeCount);
				}
				L1_neighbors[my_Id][neighborId] = yourId;
				//printf ("me %d nIndex %d you %d\n", myId, neighborId, yourId);			
		      }
		}
		else
		{
			printf("Warning ");
			// Add : My Neighbor Count < neigborCount
			int neighborId = 0;
			for(int yourId = 0; yourId < nodeCount; yourId++)
			{
				if (medians[my_Id][yourId] > -0.5 &&  belong_to[my_Id] == belong_to[yourId]) {
					L1_neighbors[my_Id][neighborId] = yourId;
					neighborId++;
				}		
			}
		}
	
    }


	printf("\nAdd L2 nbs\n");

	for (int my_Id = 0; my_Id < nodeCount; my_Id++) {
      // verify that everybody has enough neighbors
	     int L2_NeighborCount=0;
	 
	      for (int yourId = 0; yourId < nodeCount; yourId++) 
			if (medians[my_Id][yourId] > -0.5 && yourId != my_Id && belong_to_L2[my_Id] == belong_to_L2[yourId])
					L2_NeighborCount++;

			printf("[%d,%d]",my_Id,  L2_NeighborCount);
			if (L2_NeighborCount >= neighborCount / 3)
				L2_nb_count[my_Id] = neighborCount / 3;
			else
				L2_nb_count[my_Id] = L2_NeighborCount;
		if (L2_NeighborCount >  neighborCount / 3)
		{
		      for (int neighborId = 0; neighborId < (neighborCount/3); neighborId++) {
				int yourId = my_Id;
				while (yourId == my_Id || medians[my_Id][yourId] < -0.5
			       || isL2Neighbor(my_Id,yourId) || belong_to_L2[my_Id] != belong_to_L2[yourId])
			       { // <= 0. to < -0.5
					yourId = unifRand (0, nodeCount);
				}
				L2_neighbors[my_Id][neighborId] = yourId;
				//printf ("me %d nIndex %d you %d\n", myId, neighborId, yourId);			
		      }
		}
		else
		{
			printf("Warning ");
			// Add : My Neighbor Count < neigborCount
			int neighborId = 0;
			for(int yourId = 0; yourId < nodeCount; yourId++)
			{
				if (medians[my_Id][yourId] > -0.5 &&  belong_to_L2[my_Id] == belong_to_L2[yourId]) {
					L2_neighbors[my_Id][neighborId] = yourId;
					neighborId++;
				}		
			}
		}
	
    }
    */

/*
	for (int my_Id = 0; my_Id < nodeCount; my_Id++) {
      // verify that everybody has enough neighbors
	     int local_NeighborCount=0;
	 
	      for (int yourId = 0; yourId < nodeCount; yourId++) 
			if (medians[my_Id][yourId] > -0.5 && yourId != my_Id && belong_to[my_Id] == belong_to[yourId])
					local_NeighborCount++;

			printf("[%d %d] ",my_Id,  local_NeighborCount);
			if (local_NeighborCount >= neighborCount / 2)
				local_nb_count[my_Id] = neighborCount / 2;
			else
				local_nb_count[my_Id] = local_NeighborCount;
		if (local_NeighborCount >  neighborCount / 2)
		{
		      for (int neighborId = 0; neighborId < (neighborCount/2); neighborId++) {
				int yourId = my_Id;
				while (yourId == my_Id || medians[my_Id][yourId] < -0.5
			       || isLocalNeighbor(my_Id,yourId) || belong_to[my_Id] != belong_to[yourId])
			       { // <= 0. to < -0.5
					yourId = unifRand (0, nodeCount);
				}
				local_neighbors[my_Id][neighborId] = yourId;
				//printf ("me %d nIndex %d you %d\n", myId, neighborId, yourId);			
		      }
		}
		else
		{
			printf("Warning ");
			// Add : My Neighbor Count < neigborCount
			int neighborId = 0;
			for(int yourId = 0; yourId < nodeCount; yourId++)
			{
				if (medians[my_Id][yourId] > -0.5 &&  belong_to[my_Id] == belong_to[yourId]) {
					local_neighbors[my_Id][neighborId] = yourId;
					neighborId++;
				}		
			}
		}
	
    }
*/
	
	int count=0;

	int xxx[BETA];
	for(ii=0;ii<BETA;ii++) xxx[ii] = 0;

	printf("\nNode Number for Each Cluster (Different Levels):\n");
	// Node Number for Each Cluster:
	// 21 72 32 324 2 368 210 1 3 452 6 147 60 1 3 38
	printf("(L0) %d\n", nodeCount);
	printf("(L1) ");
	for(int i=0;i<BETA;i++) 
		printf("%d ", Cluster_Num[i]);
	printf("\n");
	/*
	printf("(L2) ");
	for(int i=0;i<BETA*2;i++) 
		printf("%d ", Cluster_Num_L2[i]);
	printf("\n");
	*/

	// Starting Point

	printf("Pure Vivaldi:\n");
	for(ii=0;ii<nodeCount;ii++) node[ii].Refresh();		
	generation=0;
	while (generation < ROUNDS) 
	{
		int temp_count;
		if (generation % 1000 == 0)
			printf("%d ", generation);	
		generation++;

		 for(ii=0;ii<nodeCount;ii++) {selected[ii] = -1;}
		 for(ii=0;ii<nodeCount;ii++)
		 {			 
		 	////////////////////////////////////////////////////// 		 
		 	myId = unifRand (0, nodeCount);
			while(selected[myId] != -1)
			{				 
				myId=((myId+1) % nodeCount);
			}
			selected[myId] = 1;
			rawLatencySample = -1;
			while (rawLatencySample < -0.5)			 
			{				 
				int neighborId = unifRand (0, neighborCount);
				yourId = neighbors[myId][neighborId];
				if (yourId == -1)
				{
					bad++; printf("!!"); exit(1); continue;
				}				  
				rawLatencySample = medians[myId][yourId]; 			 
			}			 
			processSample2 (stamp, myId, yourId, rawLatencySample);		 
			perSecondOutput (stamp);
			stamp++;
		}
	}
	for(ii=0;ii<nodeCount;ii++)
		for(jj=0;jj<DIMENSIONS;jj++)
			node1[ii].vec->v[jj] = node[ii].vec->v[jj]; // Save Pure Vivaldi NC
 	printf("\n");

	printf("Hierarchical Vivaldi (Global):\n");
	for(ii=0;ii<nodeCount;ii++) node[ii].Refresh();
	generation = 0;
	while (generation < ROUNDS /2) 
	{
		if (generation % 1000 == 0)
			printf("%d ", generation);	
		generation++;

		 for(ii=0;ii<nodeCount;ii++) {selected[ii] = -1;}
		 for(ii=0;ii<nodeCount;ii++)
		 {			 
		 	////////////////////////////////////////////////////// 		 
		 	myId = unifRand (0, nodeCount);
			while(selected[myId] != -1)
			{				 
				myId=((myId+1) % nodeCount);
			}
			selected[myId] = 1;
			rawLatencySample = -1;
			while (rawLatencySample < -0.5)			 
			{				 
				// Global : global_nb_upbound
				// Local   : local_nb_count[myId]
				if (local_nb_count[myId] > 0 && (rand() % 2) == 0)
				{
					int neighborId = unifRand (0, local_nb_count[myId]);
					yourId = local_neighbors[myId][neighborId];
				}
				else
				{
					int neighborId = unifRand (0, global_nb_upbound);
					yourId = global_neighbors[myId][neighborId];
				}
				if (yourId == -1)
				{
					bad++; printf("!!"); exit(1); continue;
				}				  
				rawLatencySample = medians[myId][yourId]; 			 
			}			 
			processSample2 (stamp, myId, yourId, rawLatencySample);		 
			perSecondOutput (stamp);
			stamp++;

/*
			if (local_nb_count[myId] == 0) continue;
			rawLatencySample = -1;
			while (rawLatencySample < -0.5)			 
			{				 
				int neighborId = unifRand (0, local_nb_count[myId]);
				yourId = local_neighbors[myId][neighborId];
				if (yourId == -1)
				{
					bad++; printf("!!"); 
					printf("%d %d %d %d", generation, myId, local_nb_count[myId], neighborId);
					exit(1);
					continue;
				}				  
				rawLatencySample = medians[myId][yourId]; 			 
			}			 
			processSample2 (stamp, myId, yourId, rawLatencySample);		 
			perSecondOutput (stamp);
			stamp++;
*/
			
		}
	}


	for(ii=0;ii<nodeCount;ii++)
		for(jj=0;jj<DIMENSIONS;jj++)
			node2[ii].vec->v[jj] = node[ii].vec->v[jj]; // Save Global NC of Pharos
	printf("\n");



	printf("Hierarchical Vivaldi (Local):\n");
	for(ii=0;ii<nodeCount;ii++) node[ii].Refresh();
	// Check
	/*
	for(ii=0;ii<nodeCount;ii++)
	{
		printf("%d ", local_nb_count[ii]);
	}
	printf("\n");
	*/
	generation = 0;	
	while (generation < ROUNDS /2 ) 
	{
		int temp_count;
		if (generation % 1000 == 0)
			printf("%d ", generation);	
		generation++;

		 for(ii=0;ii<nodeCount;ii++) {selected[ii] = -1;}
		 for(ii=0;ii<nodeCount;ii++)
		 {			 
		 	////////////////////////////////////////////////////// 		 
		 	myId = unifRand (0, nodeCount);
			while(selected[myId] != -1)
			{				 
				myId=((myId+1) % nodeCount);
			}
			selected[myId] = 1;
			if (local_nb_count[myId] == 0) continue;
			rawLatencySample = -1;
			while (rawLatencySample < -0.5)			 
			{				 
				int neighborId = unifRand (0, local_nb_count[myId]);
				yourId = local_neighbors[myId][neighborId];
				if (yourId == -1)
				{
					bad++; printf("!!"); 
					printf("%d %d %d %d", generation, myId, local_nb_count[myId], neighborId);
					exit(1);
					continue;
				}				  
				rawLatencySample = medians[myId][yourId]; 			 
			}			 
			processSample2 (stamp, myId, yourId, rawLatencySample);		 
			perSecondOutput (stamp);
			stamp++;
				///////////////////////////////////////////////////////// 	 
		}
	 /////////////////////////////////////////////////////////
	}

	for(ii=0;ii<nodeCount;ii++)
		for(jj=0;jj<DIMENSIONS;jj++)
			node3[ii].vec->v[jj] = node[ii].vec->v[jj]; // Save Local NC of Pharos
	printf("\n");

//	for(ii=0;ii<nodeCount;ii++)
//		printf("%d/%d/%d,", L0_nb_count[ii], L1_nb_count[ii], L2_nb_count[ii]);

/*

  printf("3-Level Vivaldi (L0):\n");
  for(ii=0;ii<nodeCount;ii++) node[ii].Refresh();
  generation = 0; 
  while (generation < ROUNDS) 
  {
	  int temp_count;
	  if (generation % 1000 == 0)
		  printf("%d ", generation);  
	  generation++;
  
	   for(ii=0;ii<nodeCount;ii++) {selected[ii] = -1;}
	   for(ii=0;ii<nodeCount;ii++)
	   {		   
		  myId = unifRand (0, nodeCount);
		  while(selected[myId] != -1)
		  { 			   
			  myId=((myId+1) % nodeCount);
		  }
		  selected[myId] = 1;
		  if (L0_nb_count[myId] == 0) continue;
		  rawLatencySample = -1;
		  while (rawLatencySample < -0.5)		   
		  { 			   
			  int neighborId = unifRand (0, neighborCount/3);
			  yourId = L0_neighbors[myId][neighborId];
			  if (yourId == -1)
			  {
				  bad++; printf("!!"); 
				  printf("%d %d %d %d", generation, myId, L0_nb_count[myId], neighborId);
				  exit(1);
				  continue;
			  } 				
			  rawLatencySample = medians[myId][yourId]; 		   
		  } 		   
		  processSample2 (stamp, myId, yourId, rawLatencySample);	   
		  perSecondOutput (stamp);
		  stamp++;
	  }
  }
  
  for(ii=0;ii<nodeCount;ii++)
	  for(jj=0;jj<DIM;jj++)
		  node4[ii].vec->v[jj] = node[ii].vec->v[jj]; // Save L0 NC of 3-Level Vivaldi
  printf("\n");


  printf("3-Level Vivaldi (L1):\n");
  for(ii=0;ii<nodeCount;ii++) node[ii].Refresh();

  generation = 0; 
  while (generation < ROUNDS) 
  {
	  int temp_count;
	  if (generation % 1000 == 0)
		  printf("%d ", generation);  
	  generation++;
  
	   for(ii=0;ii<nodeCount;ii++) {selected[ii] = -1;}
	   for(ii=0;ii<nodeCount;ii++)
	   {		   
		  myId = unifRand (0, nodeCount);
		  while(selected[myId] != -1)
		  { 			   
			  myId=((myId+1) % nodeCount);
		  }
		  selected[myId] = 1;
		  if (L1_nb_count[myId] == 0) continue;
		  rawLatencySample = -1;
		  while (rawLatencySample < -0.5)		   
		  { 			   
			  int neighborId = unifRand (0, L1_nb_count[myId]);
			  yourId = L1_neighbors[myId][neighborId];
			  if (yourId == -1)
			  {
				  bad++; printf("!!"); 
				  printf("%d %d %d %d", generation, myId, L1_nb_count[myId], neighborId);
				  exit(1);
				  continue;
			  } 				
			  rawLatencySample = medians[myId][yourId]; 		   
		  } 		   
		  processSample2 (stamp, myId, yourId, rawLatencySample);	   
		  perSecondOutput (stamp);
		  stamp++;
	  }
  }
  
  for(ii=0;ii<nodeCount;ii++)
	  for(jj=0;jj<DIM;jj++)
		  node5[ii].vec->v[jj] = node[ii].vec->v[jj]; // Save L1 NC of 3-Level Vivaldi
  printf("\n");


  printf("3-Level Vivaldi (L2):\n");
  for(ii=0;ii<nodeCount;ii++) node[ii].Refresh();
  generation = 0; 
  while (generation < ROUNDS) 
  {
	  int temp_count;
	  if (generation % 1000 == 0)
		  printf("%d ", generation);  
	  generation++;
  
	   for(ii=0;ii<nodeCount;ii++) {selected[ii] = -1;}
	   for(ii=0;ii<nodeCount;ii++)
	   {		   
		  myId = unifRand (0, nodeCount);
		  while(selected[myId] != -1)
		  { 			   
			  myId=((myId+1) % nodeCount);
		  }
		  selected[myId] = 1;
		  if (L2_nb_count[myId] == 0) continue;
		  rawLatencySample = -1;
		  while (rawLatencySample < -0.5)		   
		  { 			   
			  int neighborId = unifRand (0, L2_nb_count[myId]);
			  yourId = L0_neighbors[myId][neighborId];
			  if (yourId == -1)
			  {
				  bad++; printf("!!"); 
				  printf("%d %d %d %d", generation, myId, L2_nb_count[myId], neighborId);
				  exit(1);
				  continue;
			  } 				
			  rawLatencySample = medians[myId][yourId]; 		   
		  } 		   
		  processSample2 (stamp, myId, yourId, rawLatencySample);	   
		  perSecondOutput (stamp);
		  stamp++;
	  }
  }
  
  for(ii=0;ii<nodeCount;ii++)
	  for(jj=0;jj<DIM;jj++)
		  node6[ii].vec->v[jj] = node[ii].vec->v[jj]; // Save L0l NC of 3-Level Vivaldi
  printf("\n");  

*/

  // Generation Statistical
  printf("Finish Simulation, Total Time: %d\n", generation);
  // make sure to ignore anybody who never did anything

/*

 // Print the NC of pure Vivaldi //
  for(ii=0;ii<nodeCount;ii++) 
	  
  {
			printf("%d : ", ii);
			for (int iii = 0; iii < DIMENSIONS; iii++) {
				  printf ("%.1f ", node[ii].vec->v[iii]);
			}
			printf("\n");
  
  }

*/

/*

  if (coordFP != NULL || nodeFP != NULL) {
    for (int i = 0; i < nodeCount; i++) {
//      if (node[i].lastUpdateTo > 0) // Have been cited!
	if (period[i] > 0)
	 { 
        if (coordFP != NULL) {
          fprintf (coordFP, "%d ", i);
          node[i].printCoord (coordFP);
          fprintf (coordFP, "\n");
        }

        if (nodeFP != NULL) {
          fprintf (nodeFP, "%d ", i);
          node[i].printStat (nodeFP);
          fprintf (nodeFP, "\n");
        }
      }
    }
  }

  */

  
  int vv;
  for(vv=0;vv<2;vv++)
   {


  if (vv == 0){
  	for(int ii=0;ii<nodeCount;ii++) {period[ii] = 1;node[ii] = node1[ii];} // Point to
  	relative_fp = fopen("hv_vivaldi16.txt", "w");
	distribution_relative_fp = fopen("hv_vivaldi16_dis.txt", "a");
	rrl_fp = fopen("hv_vivaldi16_rrl.txt", "w");
	rrl_k_fp = fopen("hv_vivaldi16_rrl_k.txt", "w");
	cnl_fp = fopen("hv_vivaldi16_cnl.txt", "w");
	nnl_fp = fopen("hv_vivaldi16_nnl.txt", "w");
	ls_fp = fopen("hv_vivaldi16_ls.txt", "w");
	range_accuracy_fp = fopen("hv_vivaldi16_ra.txt", "w");
	reverse_range_accuracy_fp = fopen("hv_vivaldi16_rra.txt", "w");
	output_load_fp = fopen("hv_vivaldi16_load.txt", "w");	
  }
	if (vv == 1){
  	for(int ii=0;ii<nodeCount;ii++) {period[ii] = 1;node[ii] = node2[ii];} // Point to
	relative_fp = fopen("hv_hv16.txt", "w");
 	  distribution_relative_fp = fopen("hv_hv16_dis.txt", "a");
	  rrl_fp = fopen("hv_hv16_rrl.txt", "w");
	  rrl_k_fp = fopen("hv_hv16_rrl_k.txt", "w");
	  cnl_fp = fopen("hv_hv16_cnl.txt", "w");
	  nnl_fp = fopen("hv_hv16_nnl.txt", "w");
	  ls_fp = fopen("hv_hv16_ls.txt", "w");
	  range_accuracy_fp = fopen("hv_hv16_ra.txt", "w");
	  reverse_range_accuracy_fp = fopen("hv_hv16_rra.txt", "w");
	  output_load_fp = fopen("hv_hv16_load.txt", "w");  
	}
	/*
	if (vv == 2){
  	for(int ii=0;ii<nodeCount;ii++) {period[ii] = 1;node[ii] = node4[ii];} // Point to
	relative_fp = fopen("hv_3v16.txt", "w");
 	  distribution_relative_fp = fopen("hv_3v16_dis.txt", "w");
	  rrl_fp = fopen("hv_3v16_rrl.txt", "w");
	  rrl_k_fp = fopen("hv_3v16_rrl_k.txt", "w");
	  cnl_fp = fopen("hv_3v16_cnl.txt", "w");
	  nnl_fp = fopen("hv_3v16_nnl.txt", "w");
	  ls_fp = fopen("hv_3v16_ls.txt", "w");
	  range_accuracy_fp = fopen("hv_3v16_ra.txt", "w");
	  reverse_range_accuracy_fp = fopen("hv_3v16_rra.txt", "w");
	  output_load_fp = fopen("hv_3v16_load.txt", "w");  
	}
	*/
  	
  int cc=0;

  
//  if (reFP != NULL || sumreFP != NULL) {



	  float median = 0;
	  float sys_relativeErrorAB = 0;
	  float app_relativeErrorAB = 0;
	  float sys_distanceAB;
	  int actual_nearest;
	  int computed_nearest;
	  float rrl[MAXNODE];
	  float rrl_k[MAXNODE];
	  float cnl;

	  int nearest_good=0;
	  int total_alive = 0;

	  float min_actual;
	  float min_by_nc;
	  int ML[MAX_RANGE];
	  int ML_PL[MAX_RANGE];
	  int PL[MAX_RANGE];
  	  int ii, jj;

	  int actual_sort[MAXNODE];
	  float actual_dist[MAXNODE];  
	  int predicted_sort[MAXNODE];
	  float predicted_dist[MAXNODE];
	  int min_by_nc_id;	  

	  float min_fit_error;


	  float total_RE = 0.0, total_RRL = 0.0, total_CNAE = 0.0;
	  int total_RE_num = 0, total_RRL_num = 0, total_CNAE_num = 0;
	

   	  double stress_A = 0, stress_B = 0;
	  for(ii=0;ii<MAX_RANGE;ii++)
	  {
	  	ML[ii] = 0; ML_PL[ii] = 0; PL[ii] = 0;
	  }

//          for(ii=0;ii<nodeCount;ii++) period[ii] = 1;
	
	    for (int myId = 0; myId < nodeCount; myId++) {
		if (period[myId] > 0) {
			//      if (node[myId].lastUpdateTo > 0) {
		        vector<double> errors;
			min_actual = 10000;
			min_by_nc = 10000;
			float sys_dist[MAXNODE];

		        for (int yourId = 0; yourId < nodeCount; yourId++) {
				   sys_dist[yourId] = -1;
				   rrl[myId] = -1;
			          if (yourId != myId) 
				  {
					  if (period[yourId] > 0)
					  {
						median = 0;
						if (medians != NULL) {
							  median = medians[myId][yourId];
						}
						if (median<0)	continue;
						if (vv == 0)
						{
		 			              sys_distanceAB =  node1[myId].vec->getEuclideanDistance (node1[yourId].vec);

						}
						else if (vv == 1)
						{						
    			            sys_distanceAB =  node2[myId].vec->getEuclideanDistance (node2[yourId].vec);
							if (belong_to[myId] == belong_to[yourId])
								sys_distanceAB =  node3[myId].vec->getEuclideanDistance (node3[yourId].vec);
						}		
						/*
						else if (vv == 2)
						{
    			            sys_distanceAB =  node4[myId].vec->getEuclideanDistance (node4[yourId].vec);
							if (belong_to[myId] == belong_to[yourId])
								sys_distanceAB =  node5[myId].vec->getEuclideanDistance (node5[yourId].vec);
							if (belong_to_L2[myId] == belong_to_L2[yourId])
								sys_distanceAB =  node6[myId].vec->getEuclideanDistance (node6[yourId].vec);
						}
						*/

						stress_A += (median - sys_distanceAB) * (median - sys_distanceAB);
						stress_B += median * median;
						
//						printf("%f ", sys_distanceAB);
					       sys_dist[yourId] = sys_distanceAB;
						// median & sys_distanceAB
						if ((int)(median) < MAX_RANGE)
						{
							for(ii=(int)(median);ii<MAX_RANGE;ii++) ML[ii]++;
						}
						if ((int)(sys_distanceAB) < MAX_RANGE)
						{
							for(ii=(int)(sys_distanceAB);ii<MAX_RANGE;ii++) PL[ii]++;
						}
						if (median > sys_distanceAB)
						{
							if ((int)(median) < MAX_RANGE)
								for(ii=(int)(median);ii<MAX_RANGE;ii++) ML_PL[ii]++;
						}
						else
						{
							if ((int)(sys_distanceAB) < MAX_RANGE)
								for(ii=(int)(sys_distanceAB);ii<MAX_RANGE;ii++) ML_PL[ii]++;
						}
		 			       if (median < sys_distanceAB)
				             		sys_relativeErrorAB = fabs(sys_distanceAB - median) / median;
			     		  	else
				                      sys_relativeErrorAB = fabs(sys_distanceAB - median) / sys_distanceAB;		  	

						//////////////////////////////////////////////////////
						total_RE+= fabs(sys_distanceAB - median) / median;
//						total_RE+= sys_relativeErrorAB;
						total_RE_num++;
						//////////////////////////////////////////////////////
						
						if (median<min_actual)	min_actual = median;
						if (sys_distanceAB < min_by_nc) 
							{
							min_by_nc = sys_distanceAB;
							min_by_nc_id = yourId;
							}

				              if (reFP != NULL) {
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//							 RE in Detail
//							 fprintf (reFP, "%d %d re %f d %f l %f \n", myId, yourId, sys_relativeErrorAB, sys_distanceAB, median);
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* 2008-08-19
							 fprintf(relative_fp, "%f ", sys_relativeErrorAB); 
							if (DIMENSIONS == 7)
								 fprintf(distribution_relative_fp, "%f %f ", median,  fabs(sys_distanceAB - median) / median); 
								 */

				              }
			       	     }
			          }		  
		        }
			if (((int)min_actual) == ((int)medians[myId][min_by_nc_id])) nearest_good++;
			total_alive++;
//			fprintf(nnl_fp, "%f ", (min_by_nc - min_actual)/min_actual); // closest neighbor absolute error


			fprintf(nnl_fp, "%f ", medians[myId][min_by_nc_id] - min_actual); // closest neighbor absolute error

			//////////////////////////////
			total_CNAE+=  medians[myId][min_by_nc_id] - min_actual;
			total_CNAE_num++;

			//////////////////////////////


			fprintf(ls_fp, "%f ", (medians[myId][min_by_nc_id] + 10) / (min_actual + 10)); // closest neighbor absolute error
			int tt1=0, tt2=0;
			int nearest_k = 1; // Top K nearest neighbors
			for(int yourId1=0;yourId1<nodeCount;yourId1++)
			{
				if (period[yourId1] < 0 || medians[myId][yourId1] < 0 || yourId1 == myId) continue;
				for(int yourId2=0;yourId2<yourId1;yourId2++)
				{
					if (period[yourId2] < 0 || medians[myId][yourId2] < 0 || yourId2 == myId) continue;
//				if (yourId1 == yourId2) continue;
					tt1++;
					if ((medians[myId][yourId1] > medians[myId][yourId2] && sys_dist[yourId1] > sys_dist[yourId2]) ||
					(medians[myId][yourId1] < medians[myId][yourId2] && sys_dist[yourId1] < sys_dist[yourId2]) ||
					(medians[myId][yourId1] == medians[myId][yourId2] && sys_dist[yourId1] == sys_dist[yourId2]))
						tt2++;
				}
			}

			for(ii=0;ii<nodeCount;ii++)
			{
//				actual_sort[ii] = ii; 
				actual_dist[ii] = medians[myId][ii];
				if (actual_dist[ii] < 0 || myId == ii) actual_dist[ii] = 100000;
//				predicted_sort[ii] = ii;


// Marked: 2007-01-19
//				predicted_dist[ii] =  node[myId].vec->getEuclideanDistance (node[ii].vec);




// Added : 2007-01-19
				if (vv == 0)
					predicted_dist[ii] =  node1[myId].vec->getEuclideanDistance (node1[ii].vec);
				else if (vv == 1)
				{						
					predicted_dist[ii] =  node2[myId].vec->getEuclideanDistance (node2[ii].vec);
// 			              sys_distanceAB =  node1[myId].vec->getEuclideanDistance (node1[ii].vec);
					if (belong_to[myId] == belong_to[ii])
						predicted_dist[ii] =  node3[myId].vec->getEuclideanDistance (node3[ii].vec);
				}
				/*
				else
				{						
					predicted_dist[ii] =  node4[myId].vec->getEuclideanDistance (node4[ii].vec);

					if (belong_to[myId] == belong_to[ii])
						predicted_dist[ii] =  node5[myId].vec->getEuclideanDistance (node5[ii].vec);
					if (belong_to_L2[myId] == belong_to_L2[ii])
						predicted_dist[ii] =  node6[myId].vec->getEuclideanDistance (node6[ii].vec);
						
				}*/



				//////////////////////////////!!!!!!!!!!!!!!!!!
				if (myId == ii) predicted_dist[ii] = 100000;
			}
/*
			int tttt, hit = 0;
			for(ii=0;ii<nodeCount-1;ii++)
				for(jj=ii+1;jj<nodeCount;jj++)
					{
						if((period[ii] < 0 && period[jj] > 0) ||
							((period[ii] > 0 && period[jj] > 0) && actual_dist[ii] > actual_dist[jj]))
							{
								tttt = actual_sort[ii]; actual_sort[ii] = actual_sort[jj]; actual_sort[jj] = tttt;			
							}
					}
			for(ii=0;ii<nodeCount-1;ii++)
				for(jj=ii+1;jj<nodeCount;jj++)
					{
					if((period[ii] < 0 && period[jj] > 0) ||
						((period[ii] > 0 && period[jj] > 0) && predicted_dist[ii] > predicted_dist[jj]))
						{
							tttt = predicted_sort[ii]; predicted_sort[ii] = predicted_sort[jj]; predicted_sort[jj] = tttt;			
						}
					}
					*/
			/*
			for(ii=0;ii<nearest_k;ii++)
			{
				for(jj=0;jj<nearest_k;jj++)
				{
					if (actual_sort[ii] == predicted_sort[jj])
					{
						hit++;
						break;
					}
				}
			}
			*/
			float x1, x2, xmin;
			int min_number;
			int processed_actual[MAXNODE], processed_predicted[MAXNODE];
			for(ii=0;ii<nodeCount;ii++) {processed_actual[ii] = -1;processed_predicted[ii] = -1;}
			for(ii=0;ii<nearest_k;ii++)
			{
				// x1 -> searching
				xmin = 10000; min_number = 0;
				for(jj=0;jj<nodeCount;jj++)
					if (actual_dist[jj] < xmin && processed_actual[jj] == -1)
					{
						xmin = actual_dist[jj]; min_number = jj;
					}
				x1 = xmin;
				processed_actual[min_number] = 1;			
				// x2 -> searching
				xmin = 10000; min_number = 0;
				for(jj=0;jj<nodeCount;jj++)
					if (predicted_dist[jj] < xmin && processed_predicted[jj] == -1)
					{
						xmin = predicted_dist[jj]; min_number = jj;
					}
				x2 = xmin;
				processed_predicted[min_number] = 1;	
			}
			for(ii=0;ii<nearest_k;ii++)
			{
				// x1 -> searching
				xmin = 10000; min_number = 0;
				for(jj=0;jj<nodeCount;jj++)
					if (actual_dist[jj] < xmin && processed_actual[jj] == 1)
					{
						xmin = actual_dist[jj]; min_number = jj;
					}
				x1 = xmin;
				processed_actual[min_number] = 2;			
				// x2 -> searching
				xmin = 10000; min_number = 0;
				for(jj=0;jj<nodeCount;jj++)
					if (actual_dist[jj] < xmin && processed_predicted[jj] == 1)
					{
						xmin = actual_dist[jj]; min_number = jj;
					}
				x2 = xmin;
				processed_predicted[min_number] = 2;	
				fprintf(rrl_k_fp, "%f ", x2-x1); // closest neighbor absolute error
			}

			/////////////////////////
			// Do Sorting!!
			/////////////////////////
			int hit = 0;
			for(ii=0;ii<nodeCount;ii++)
				if (processed_actual[ii] > -1 && processed_predicted[ii] > -1)
					hit++;
			rrl[myId] = (float)(tt1-tt2) / tt1;
			rrl_k[myId] = (float)(nearest_k-hit) / nearest_k;
			fprintf(rrl_fp, "%f ", rrl[myId]); // closest neighbor absolute error
			fprintf(rrl_k_fp, "%f ", rrl_k[myId]); // closest neighbor absolute error



			//////////////////////////////
			total_RRL+=  rrl[myId];
			total_RRL_num++;

			//////////////////////////////


//			printf("[%f] ", rrl_k[myId]); // closest neighbor absolute error

			if (sumreFP != NULL) {
			  double fifty = percentile (errors, .5);
			  double eighty = percentile (errors, .8);
			  double ninetyfive = percentile (errors, .95);
			  fprintf (sumreFP, "%d %f %f %f\n", myId,
					   fifty, eighty, ninetyfive);
			}

      		}
	}

	if (vv == 0)
		printf("WhereIs (1) -> Vivaldi  :\n");
	else if (vv == 1)
		printf("WhereIs (2) -> Pharos :\n");
/*
	else if (vv == 2)
		printf("WhereIs (3) -> 3Vivaldi :\n");
		*/

	int tree_root_node = rand() % nodeCount;


	MST(vv, tree_root_node);
	ESM(vv, tree_root_node);
	LGK (vv, tree_root_node);
	SRV(vv);
//	Unstructured(vv);

	for(ii=0;ii<MAX_RANGE;ii++)
	{
		if (ML[ii] == 0)
			fprintf(range_accuracy_fp, "%f ", 0);
		else
			fprintf(range_accuracy_fp, "%f ", ((float)ML_PL[ii]) / ML[ii]);
		if (PL[ii] == 0)
			fprintf(reverse_range_accuracy_fp, "%f ", 0);
		else
			fprintf(reverse_range_accuracy_fp, "%f ", ((float)ML_PL[ii]) / PL[ii]);
	}
	
    printf("CNL = %f\n", (float)(total_alive-nearest_good)*100/total_alive);
	printf("Stress = %f\n", sqrt(stress_A / stress_B));
	
	printf("Average RE = %f\n", total_RE / total_RE_num);
	printf("Average RRL = %f\n", total_RRL / total_RRL_num);
	printf("Average CNAE = %f\n", total_CNAE / total_CNAE_num);
	
	fprintf(cnl_fp, "CNL = %f\n", (float)(total_alive-nearest_good)*100/total_alive);
	fprintf(range_accuracy_fp, "\n");
	fprintf(reverse_range_accuracy_fp, "\n");

//  }



  fclose(relative_fp);
  fclose(rrl_fp);	
  fclose(cnl_fp);
  fclose(nnl_fp);
  fclose(ls_fp);
  fclose(range_accuracy_fp);
  fclose(output_load_fp);
  fclose(distribution_relative_fp);
   	}



  	}

/*
	  // Binning //
	  FILE *binning_nnl_fp = fopen("hv_binning16_nnl.txt", "w");
	  FILE *binning_ls_fp = fopen("hv_binning16_LS.txt", "w");
	  double temp_dist[BETA];
	  double temp_dist_k;
	  int temp_index;
	  int server_selection[MAXNODE];
  
	  for(int ii=0;ii<BETA;ii++)
		  first_beta[ii] = rand() % nodeCount;
  
	  for(int ii=0;ii<nodeCount;ii++)
	  {
		  for(int jj=0;jj<BETA;jj++)
		  {
			  binning[ii][jj] = jj;
			  temp_dist[jj] = medians[ii][first_beta[jj]];
			  if (temp_dist[jj] < 0)
				  temp_dist[jj] = 10000;
		  }
		  
		  // Sort
  
		  for(int jj=0;jj<BETA-1;jj++)
			  for(int kk = jj+1;kk<BETA;kk++)
				  {
					  if (temp_dist[jj] > temp_dist[kk])
					  {
						  temp_dist_k = temp_dist[jj];
						  temp_dist[jj] = temp_dist[kk];
						  temp_dist[kk] = temp_dist_k;
						  temp_index = binning[ii][jj];
						  binning[ii][jj] = binning[ii][kk];
						  binning[ii][kk] = temp_index; 			  
						  
					  }
				  }
  //	  printf("%d :", ii);
  //	  for(int jj=0;jj<BETA;jj++)
  //		  printf("%d ", binning[ii][jj]);
  //	  printf("\n");
	  }
	  // Binning Based Server Selection
	  for(int ii=0;ii<nodeCount;ii++) server_selection[ii] = 1;
	  for(int ii=0;ii<nodeCount;ii++)	  
	  {
		  double x1 = 10000, x2 = 10000;
		  int temp_closest = 0, temp_closest_score = 0;
		  // Closest
		  for(int jj=0;jj<nodeCount;jj++)
		  {
			  if (jj != ii && medians[ii][jj] >= 0 && medians[ii][jj] < x1 && server_selection[jj] == 1)
				  x1 = medians[ii][jj];
		  }
  //	  printf("<%lf ", x1);
		  // Closest by Binning
		  int score1, score2;
		  int init = rand() % nodeCount;
		  for(int jj=0;jj<nodeCount;jj++)
		  {
			  // Compare jj <-> temp_closest
			  if (medians[ii][(jj+init)%nodeCount] < 0)
				  continue;
			  if (((jj+init)%nodeCount)== ii)
				  continue;
			  if (server_selection[(jj+init)%nodeCount] != 1)
				  continue;
			  score1 = 0;
			  for(int kk = 0;kk<BETA;kk++)
			  {
				  if (binning[ii][kk] == binning[(jj+init)%nodeCount][kk])
					  score1++;
  //			  else
  //				  break;
			  }
			  if (score1 >= temp_closest_score)
			  {
				  temp_closest_score = score1;
				  temp_closest = jj;
			  } 		  
		  }
		  x2 = medians[ii][(temp_closest+init)%nodeCount];
  //	  printf("%lf> ", x2);
		  fprintf(binning_nnl_fp, "%f ", x2 - x1); // closest neighbor absolute error
		  fprintf(binning_ls_fp, "%f ", (x2 + 10)/(x1+10)); // latency stretch
  
	  }
	  fclose(binning_nnl_fp);
	  fclose(binning_ls_fp);
*/	


  

  /*

  if (outputFormat == 'l' ||
  outputFormat == 'a' ||
  outputFormat == 'p') {

  for (int myId = 0; myId < nodeCount; myId++) {
  float loss = 0;
  float sum = 0; 

  for (int nodeA = 0; nodeA < nodeCount; nodeA++) {
  for (int nodeB = nodeA+1; nodeB < nodeCount; nodeB++) {

  if (myId != nodeA && myId != nodeB &&
  medians[myId][nodeA] > 0 && medians[myId][nodeB] > 0 &&
  node[myId].lastUpdateTo > 0 && node[nodeA].lastUpdateTo > 0 &&
  node[nodeB].lastUpdateTo > 0) {


  float distanceA = 
  node[myId].vec->getEuclideanDistance (node[nodeA].vec);
  float distanceB = 
  node[myId].vec->getEuclideanDistance (node[nodeB].vec);

  float medianA = medians[myId][nodeA];
  float medianB = medians[myId][nodeB];

  float medianDifference = fabs(medianA-medianB);

  if (debug) {
  printf ("A %d B %d dA %f sA %f dB %f sB %f\n",
  nodeA, nodeB,
  distanceA, medianA,
  distanceB, medianB);
  }

  switch (outputFormat) {
  case 'l':
  sum++;
  if (medianA > medianB) {
  if (distanceA < distanceB)
  loss++;
  } else if (medianB > medianA) {
  if (distanceB < distanceA)
  loss++;
  }
  break;
                
  case 'a':
  sum += medianDifference;
  if (medianA > medianB) {
  if (distanceA < distanceB) 
  loss += medianDifference;
  } else if (medianB > medianA) {
  if (distanceB < distanceA) 
  loss += medianDifference;
  }
  break;

  case 'p':
  if (medianA > medianB) {
  if (distanceA < distanceB) {
  loss += medianDifference;
  sum += medianA;
  }
  } else if (medianB > medianA) {
  if (distanceB < distanceA) {
  loss += medianDifference;
  sum += medianB;
  }
  }
  break;
  }
  }
  }
  }
  float lossPct = loss/sum;
  printf ("%d %f\n", myId, lossPct);

  }
  }

  */
delete [] node;

  if (medians != NULL) {
    for (int myId = 0; myId < nodeCount; myId++) {
      delete [] medians[myId];
    }
    delete [] medians;
  }

  return 0;
}


FILE* openOutputFile (char *prefix, char *suffix) {
  FILE *fp;
  char *filename;
  int fileLength = strlen(prefix)+strlen(suffix)+1;

  filename = new char [fileLength];
  memset (filename,0,fileLength);
  sprintf (filename, "%s%s", prefix, suffix);
  if ((fp = fopen(filename, "w")) == NULL) {
    printf("%s: file open error.\n", filename);
    exit (-1);
  }
  delete [] filename;
  return fp;
}

void openOutputFile (FILE *&fp, char *prefix, char *suffix, int round) {
  char *filename;
  int fileLength = strlen(prefix)+strlen(suffix)+32;

  filename = new char [fileLength];
  memset (filename,0,fileLength);
  sprintf (filename, "%s%05d%s", prefix, round, suffix);
  if ((fp = fopen(filename, "w")) == NULL) {
    printf("%s: file open error.\n", filename);
    exit (-1);
  }
  delete [] filename;
}

void closeOutputFile (FILE *fp) {
  fclose (fp);
  fp = NULL;
}

bool isNeighbor (int myId, int yourId) {
  for (int neighborId = 0; neighborId < neighborCount; neighborId++) {
    if (neighbors[myId][neighborId] == yourId) {
      return true;
    }
  }
  return false;
}
bool isGlobalNeighbor (int myId, int yourId) {
  for (int neighborId = 0; neighborId < (neighborCount / 2); neighborId++) {
    if (global_neighbors[myId][neighborId] == yourId) {
      return true;
    }
  }
  return false;
}
bool isLocalNeighbor (int myId, int yourId) {
  for (int neighborId = 0; neighborId < (neighborCount / 2); neighborId++) {
    if (local_neighbors[myId][neighborId] == yourId) {
      return true;
    }
  }
  return false;
}


bool isL0Neighbor (int myId, int yourId) {
  for (int neighborId = 0; neighborId < (neighborCount / 3); neighborId++) {
    if (L0_neighbors[myId][neighborId] == yourId) {
      return true;
    }
  }
  return false;
}

bool isL1Neighbor (int myId, int yourId) {
  for (int neighborId = 0; neighborId < (neighborCount / 3); neighborId++) {
    if (L1_neighbors[myId][neighborId] == yourId) {
      return true;
    }
  }
  return false;
}

bool isL2Neighbor (int myId, int yourId) {
  for (int neighborId = 0; neighborId < (neighborCount / 3); neighborId++) {
    if (L2_neighbors[myId][neighborId] == yourId) {
      return true;
    }
  }
  return false;
}



bool is_base_layer_Neighbor (int mybaseId, int yourbaseId) {
  for (int neighborId = 0; neighborId < neighborCount; neighborId++) {
    if (base_layer_nb[mybaseId][neighborId] == yourbaseId) {
      return true;
    }
  }
  return false;
}


int search_base_layer(int yourId)
{
	int ii;
	for(ii=0;ii<BASESIZE;ii++)
		if (yourId == base_layer_nodes[ii])
			return ii;
	return -1;
}

void perSecondOutput (int stamp) {
  if (stamp > lastFrameStamp + FRAME_INTERVAL) {
    lastFrameStamp += FRAME_INTERVAL;
    if (secFP) {
      double sysRawRE_50 = percentile (sec_sysRawRE_list, .5);
      double sysRawRE_80 = percentile (sec_sysRawRE_list, .8);
      double sysRawRE_95 = percentile (sec_sysRawRE_list, .95);

      double appRawRE_50 = percentile (sec_appRawRE_list, .5);
      double appRawRE_80 = percentile (sec_appRawRE_list, .8);
      double appRawRE_95 = percentile (sec_appRawRE_list, .95);

      fprintf (secFP, "%d %d %.3f  %.3f  %.3f  %.3f  %.3f  %.3f  %.3f  %.3f\n",
	       lastFrameStamp/FRAME_INTERVAL, sec_appRawRE_list.size(),
	       sysRawRE_50, sysRawRE_80, sysRawRE_95, sec_sysDD_sum, 
	       appRawRE_50, appRawRE_80, appRawRE_95, sec_appDD_sum);
      sec_sysRawRE_list.clear ();
      sec_appRawRE_list.clear ();
      sec_sysDD_sum = 0;
      sec_appDD_sum = 0;
    }

    if (strchr(outputFormat,'g') != NULL) {
      FILE *gnuplotFP = new FILE;;
      openOutputFile (gnuplotFP, outputPrefix, ".gp", lastFrameStamp/FRAME_INTERVAL);

      for (int i = 0; i < nodeCount; i++) {
	if (node[i].lastUpdateTo > 0) {
	  fprintf (gnuplotFP, "%d ", i);
	  node[i].printCoord (gnuplotFP);
	  node[i].printAppCoord (gnuplotFP);
	  fprintf (gnuplotFP, "\n");
	}
      }
      closeOutputFile (gnuplotFP);

    }
  }
}

void MST(int vv, int tree_root_node)
{
	// MST Contruction //
	// Evaluate the Tree Cost //
	int ii,jj,kk;
	// Prim Algorithm
	double Prim_cost, Actual_cost;
	for(ii=0;ii<nodeCount;ii++)
		Prim_table[ii] = -1;
	Prim_cost = 0;
	Actual_cost = 0;
	for(ii=0;ii<nodeCount;ii++)
	{
		if (ii==0) // first round
		{
			Prim_table[tree_root_node] = 1;
		}
		else
		{
			double min_dist = 10000;
			double temp_dist;
			int temp_jj, temp_kk;
			for(jj=0;jj<nodeCount;jj++)
			{
				if (Prim_table[jj] != 1)
					continue;
				for(kk=0;kk<nodeCount;kk++)
				{
					if (medians[jj][kk] < 0)
						continue;
					if (Prim_table[kk] == 1)
						continue;

					if (vv == 0)
					{
						temp_dist =	node1[jj].vec->getEuclideanDistance (node1[kk].vec);
					
					}
					else if (vv == 1)
					{						
						temp_dist =  node2[jj].vec->getEuclideanDistance (node2[kk].vec);
						if (belong_to[jj] == belong_to[kk])
							temp_dist =  node3[jj].vec->getEuclideanDistance (node3[kk].vec);
					}			
					if (temp_dist < min_dist)
					{
						min_dist = temp_dist;
						temp_jj = jj;
						temp_kk = kk;
					}
				}
			}
			Prim_table[temp_kk] = 1;
			Prim_cost += min_dist;
			Actual_cost+= medians[temp_jj][temp_kk];
		}
	}
	printf("MST Cost = %f, Actual Cost = %f\n", Prim_cost, Actual_cost);
	///////////////////////

}

void ESM(int vv, int tree_root_node)
{
	// Modified ESM //
	double ESM_cost, Actual_ESM_cost;
	int ESM_current_node;
	int ii,jj,kk;
	for(ii=0;ii<nodeCount;ii++) ESM_table[ii] = -1;
	ESM_cost = 0; Actual_ESM_cost = 0;

	for(ii=0;ii<nodeCount;ii++)
	{
		if (ii== 0)
		{
			ESM_table[tree_root_node] = 1;
		}
		else
		{
			int have_connection;
			do
			{
				have_connection = 0;
				ESM_current_node = rand() % nodeCount;
				for(jj=0;jj<nodeCount;jj++)
				{
					if (ESM_table[jj] == 1 && medians[ESM_current_node][jj] > 0)
					{
						have_connection = 1;
						break;
					}
				}				
			} while(have_connection == 0);
			// ESM_current_node is joining
			ESM_table[ESM_current_node] = 1;
			// Create the link
			if (ii<30) // First 30 Nodes
			{
				double min_dist = 10000, temp_dist; int temp_jj = -1;
				for(jj=0;jj<nodeCount;jj++)
				{
					if (ESM_table[jj] != 1)	continue;		
					if (ESM_current_node == jj) continue;
					if (medians[ESM_current_node][jj] < 0) continue; 
					if (vv == 0)
					{
						temp_dist =	node1[ESM_current_node].vec->getEuclideanDistance (node1[jj].vec);					
					}
					else if (vv == 1)
					{						
						temp_dist =  node2[ESM_current_node].vec->getEuclideanDistance (node2[jj].vec);
						if (belong_to[ESM_current_node] == belong_to[jj])
							temp_dist =  node3[ESM_current_node].vec->getEuclideanDistance (node3[jj].vec);
					}							
					if (temp_dist < min_dist)
					{
						min_dist = temp_dist; temp_jj = jj;
					}
				}
				if (temp_jj != -1)
				{
					ESM_cost += min_dist;
					Actual_ESM_cost += medians[ESM_current_node][temp_jj];
//					printf("%.2f ", min_dist);
				}
				else
				{
					printf("warning at %d\n", ii);
					break;
				}
			}
			else
			{
				double min_dist = 10000, temp_dist; int temp_jj = -1;
				for(jj=0;jj<nodeCount;jj++)
					ESM_thirty_candidate[jj] = -1;
				for(int ESM_try = 0; ESM_try<30; ESM_try++)
				{
					jj = rand() % nodeCount;
					int ori_jj = jj, have_candidate = 1;
					while (ESM_table[jj] != 1
						|| ESM_thirty_candidate[jj] == 1
						|| medians[ESM_current_node][jj] < 0
						|| ESM_current_node == jj)
						{
							jj=(jj+1) % nodeCount;					
							if (jj == ori_jj) 
							{
								break;
								have_candidate = 0;
							}
						}
					if (have_candidate == 0)
						break;
					
					ESM_thirty_candidate[jj] = 1;

					if (medians[ESM_current_node][jj] < 0)
						continue;
					//if (ESM_table[jj] != 1)	continue;					
					if (vv == 0)
					{
						temp_dist =	node1[ESM_current_node].vec->getEuclideanDistance (node1[jj].vec);					
					}
					else if (vv == 1)
					{						
						temp_dist =  node2[ESM_current_node].vec->getEuclideanDistance (node2[jj].vec);
						if (belong_to[ESM_current_node] == belong_to[jj])
							temp_dist =  node3[ESM_current_node].vec->getEuclideanDistance (node3[jj].vec);
					}							
					if (temp_dist < min_dist)
					{
						min_dist = temp_dist; temp_jj = jj;
					}
				}

				if (temp_jj != -1)
				{
					ESM_cost += min_dist;
					Actual_ESM_cost += medians[ESM_current_node][temp_jj];
//					printf("%.2f ", min_dist);
				}
				else
				{
					printf("warning at %d\n", ii);
					break;
				}

			}
		}			
	}
	///////////////////////////////////////////////////
	printf("ESM Cost = %f, Actual Cost = %f\n", ESM_cost, Actual_ESM_cost);



}
	

void LGK(int vv, int tree_root_node) 	// LGK Algorithm
{
	int ii, jj, kk;
	int LGK_order[MAXNODE];
	int LGK_father[MAXNODE];
	int LGK_belong_to[MAXNODE];
	static double predicted_matrix[MAXNODE][MAXNODE];
	int LGK_k = 2, current_node;
	int LGK_root_num = 0;
	int root;

	double total_stretch;
	int total_stretch_num;

	for(ii=0;ii<nodeCount;ii++)
		for(jj=0;jj<nodeCount;jj++)
			if (ii == jj)
				predicted_matrix[ii][jj] = -1;
			else
			{
				if (vv == 0)
				{
					predicted_matrix[ii][jj] = node1[ii].vec->getEuclideanDistance (node1[jj].vec);					
				}
				
				else if (vv == 1)
				{						
					predicted_matrix[ii][jj] =  node2[ii].vec->getEuclideanDistance (node2[jj].vec);					
					if (belong_to[ii] == belong_to[jj])
						predicted_matrix[ii][jj] =  node3[ii].vec->getEuclideanDistance (node3[jj].vec); 
				}							
			}

//	int perfect = 1;
//	do
//	{
	for(ii=0;ii<nodeCount;ii++) // Initial Value
	{
		LGK_order[ii] = -1;
		LGK_father[ii] = -1;
	}
	total_stretch = 0.0;
	total_stretch_num = 0;
				
	for(ii=0;ii<nodeCount;ii++)
	{
		if (ii==0) // first round
		{
			// Select the root
			current_node = tree_root_node;
			root = current_node;
			LGK_order[current_node] = 0;
			for(jj=0;jj<nodeCount;jj++)
				LGK_belong_to[jj] = current_node;
			LGK_belong_to[current_node] = -1;
		}
		else
		{
			for(jj=0;jj<nodeCount;jj++)
				if (LGK_order[jj] == ii)
					break;
			if (jj == nodeCount)
			{
				printf("warning1 at round %d\n", ii);
//				perfect = 0;
				break;
			}
			current_node = jj;
		}
			
		// Select k closest neighbors
		for(jj=0;jj<LGK_k;jj++)
		{
			// Select current closest neighbor
			double closest_dist = 100000;
			int closest_node = -1;
			int belong_to_count = 0;
			int temp;
			double overlay_dist = 0.0;
			for(kk=0;kk<nodeCount;kk++)
			{
				if(LGK_order[kk] > 0) continue; // already a root
				if (medians[current_node][kk] < 0) continue;	// no route to the host
				if (LGK_belong_to[kk] != current_node)
					continue;
				else
					belong_to_count++;
				if (predicted_matrix[current_node][kk] < closest_dist)
				{
					closest_dist = predicted_matrix[current_node][kk];
					closest_node = kk;
				}	
			}
			if (closest_node == -1 && belong_to_count > 0)
				printf("warning2\n");
			if (closest_node != -1)
			{
				LGK_root_num++;
				LGK_order[closest_node] = LGK_root_num;
				LGK_father[closest_node] = current_node;
				// Calculate the stretch
				temp = closest_node;
				while(temp != root)
				{
					overlay_dist+= medians[temp][LGK_father[temp]];				
					temp=LGK_father[temp];
				}
//				printf("%.2f,%.2f\t", overlay_dist, medians[closest_node][root]);
				if (medians[closest_node][root] > 0)
				{
					total_stretch+=overlay_dist / medians[closest_node][root];
					total_stretch_num ++;
				}
					

				///////////////////////////
//				printf("(%d,%d)", closest_node,current_node);
			}
		}
//		printf(" ");
		// Group into k-clusters
		for(jj=0;jj<nodeCount;jj++)
		{
			if (LGK_belong_to[jj] == current_node) // Split the orginal cluster into k groups				
			{
				int closest_sub_root = -1;
				double closest_sub_root_dist = 100000;
				
				if (LGK_father[jj] == current_node)
					LGK_belong_to[jj] = current_node;
					
				for(kk=0;kk<nodeCount;kk++)
				{
					if (LGK_father[kk] != current_node)
						continue;
					if (kk == jj)
						continue;
					////
					if (predicted_matrix[jj][kk] < closest_sub_root_dist)
					{
						if (closest_sub_root != -1 && medians[kk][jj] < 0)
							continue;
						closest_sub_root_dist = predicted_matrix[jj][kk];
						closest_sub_root = kk;
					}
					////					
				}

				LGK_belong_to[jj] = closest_sub_root;
			}
		}
		
	}
//	}while(perfect == 0);
//	printf("MST Cost = %f, Actual Cost = %f\n", Prim_cost, Actual_cost);
	///////////////////////
	printf("LGK Stretch = %.2f\n", total_stretch/total_stretch_num);
	return;
}

void SRV(int vv)
{
	  // SERV_Selection //

	  double temp_dist, min_srv_dist;
	  int temp_index;
	  int server_selection[MAXNODE];
	  double average_LS = 0, total_LS = 0, LS_num = 0;
	  int ii;
  
	  // Total SRV: 30 for PL (226), 200 for King (1740)
	  
	  for(int ii=0;ii<nodeCount;ii++) server_selection[ii] = 1;

	  if (nodeCount == 226)
	  	for(ii=30;ii<nodeCount;ii++) server_selection[ii] = 0;
	  if (nodeCount == 1740)
	  	for(ii=200;ii<nodeCount;ii++) server_selection[ii] = 0;
	  
//  	for(ii=50;ii<nodeCount;ii++) server_selection[ii] = 0;
	int tmp_x1, tmp_x2, tmp_srv;
	for(ii=0;ii<10000;ii++)
	{
		tmp_x1 = rand() % nodeCount;		
		tmp_x2 = rand() % nodeCount;
		tmp_srv = server_selection[tmp_x1]; 
		server_selection[tmp_x1] = server_selection[tmp_x2]; 
		server_selection[tmp_x2] = tmp_srv;
		
	}

	  
	  for(int ii=0;ii<nodeCount;ii++)	  
	  {
		  double x1 = 10000, x2 = 10000;
		  int temp_closest = 0, temp_closest_score = 0;
		  if (server_selection[ii] == 1)
		  	continue;
		  // Closest
		  for(int jj=0;jj<nodeCount;jj++)
		  {
			  if (jj != ii && medians[ii][jj] >= 0 && medians[ii][jj] < x1 && server_selection[jj] == 1)
				  x1 = medians[ii][jj];
		  }
		  //	  printf("<%lf ", x1);
		  // Closest by NC
		  int score1, score2;
		  min_srv_dist = 10000;
		  for(int jj=0;jj<nodeCount;jj++)
		  {
			  // Compare jj <-> temp_closest
			  if (medians[ii][jj] < 0)
				  continue;
			  if (jj == ii)
			  	continue;
			  if (server_selection[jj] != 1)
				  continue;
			
			if (vv == 0)
			{
				temp_dist =	node1[ii].vec->getEuclideanDistance (node1[jj].vec);					
			}
			else if (vv == 1)
			{						
				temp_dist =  node2[ii].vec->getEuclideanDistance (node2[jj].vec);
				if (belong_to[ii] == belong_to[jj])
					temp_dist =  node3[ii].vec->getEuclideanDistance (node3[jj].vec);
			}							
			if (temp_dist < min_srv_dist)
			{
				min_srv_dist = temp_dist; temp_closest= jj;
			}
		  }
		  x2 = medians[ii][temp_closest];
//		  printf("%f ", x2-x1);
		  total_LS += ((x2)/(x1));
		  LS_num++;
		  
//		  fprintf(binning_ls_fp, "%f ", (x2 + 10)/(x1+10)); // latency stretch
  
	  }
	 printf("Average LS = %f\n", total_LS / LS_num);
	 return;

}

void Unstructured(int vv)
{

	int ii,jj,kk;
	double cost = 0;
	for(ii=0;ii<nodeCount;ii++)
		for(jj=0;jj<nodeCount;jj++)
			unstructured_link_matrix[ii][jj] = -1;

	for(ii=0;ii<nodeCount;ii++)
	{
		double temp_dist, min_srv_dist;
		int temp_closest;
		for(jj=0;jj<6;jj++)
		{
			
			min_srv_dist = 100000;
			for(int kk=0;kk<nodeCount;kk++)
			{
				// Compare jj <-> temp_closest
				if (medians[ii][kk] < 0)
					continue;
				if (ii==kk)
					continue;
				if (unstructured_link_matrix[ii][kk] > 0)
					continue;
	                     if (vv == 0)
				 {
					temp_dist = node1[ii].vec->getEuclideanDistance (node1[kk].vec);					
				 }
 				else if (vv == 1)
				{						
					temp_dist =  node2[ii].vec->getEuclideanDistance (node2[kk].vec);
					if (belong_to[ii] == belong_to[kk])
						temp_dist =  node3[ii].vec->getEuclideanDistance (node3[kk].vec);
				}							
				if (temp_dist < min_srv_dist)
				{
					min_srv_dist = temp_dist; temp_closest= kk;
				}
			  }
		         cost += medians[ii][temp_closest];			
			unstructured_link_matrix[ii][temp_closest] = 1;
		}
		do
		{
			kk = rand() % nodeCount;
			if (unstructured_link_matrix[ii][kk] < 0)
			{
				unstructured_link_matrix[ii][kk] = -1;
				cost += medians[ii][kk];
				break;
			}
		} while(1);
		
	}
	printf("Unstructured Overlay = %f\n", cost);
	///////////////////////
}


